"""Command handlers for:

eda pnr --tool=openroad TARGET
"""

import os
import shutil
from pathlib import Path

from opencos import util
from opencos.commands.pnr import CommandPnr
from opencos.eda_base import Tool
from opencos.files import get_source_files_paths
from opencos.tools.slang_yosys import CommandFListSlangYosys
from opencos.utils import docker_checks
from opencos.utils.docker_checks import docker_ok


class ToolOpenRoad(Tool):
    """Tool class for: eda COMMAND --tool=openroad [args] TARGETS"""

    _TOOL = "openroad"
    _EXE = (
        ""  # This is OK to leave empty str, will use config.tools.TOOL.requires-docker
    )
    _URL = "https://theopenroadproject.org/"

    DEFAULT_DOCKER_IMAGE = "ghcr.io/cognichip/openroad-flow:latest"

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        if not docker_ok():
            # docker is required for this tool.
            self.error(
                f'For --tool={self._TOOL}, you must have functioning "docker"',
                "see --help for args --docker-run and --docker-image",
            )
            self._VERSION = "UNKNOWN"
            return self._VERSION

        self.get_set_default_docker_image()

        if not docker_checks.check_image_available(image=self.DEFAULT_DOCKER_IMAGE):
            # Be nice to eda --help, don't call docker run to get versioning if
            # the image hasn't been pulled.
            self._VERSION = "IMAGE-NOT-PULLED"
            return self._VERSION

        # this is a little customized, b/c we don't really have an Open Road
        # "version", all we have is a git tag or git hash
        git_info_str = (
            "cd /tools/OpenROAD-flow-scripts && echo "
            "$(git log -1 --format=%cs) "
            "$(git rev-parse --abbrev-ref HEAD) "
            "$(git rev-parse --short HEAD)"
        )
        env_dict = docker_checks.get_docker_env(cmd_des_obj=self)
        version_ret = docker_checks.run_docker_for_version_info(
            image=self.DEFAULT_DOCKER_IMAGE,
            env_dict=env_dict,
            commands=["-c", git_info_str],
        )
        if version_ret:
            self._VERSION = ".".join(version_ret.strip().split()).replace("..", ".")
        else:
            self._VERSION = "UNKNOWN"

        return self._VERSION


class CommandPnrOpenRoad(CommandPnr, ToolOpenRoad):
    """Command handler for: eda pnr --tool=openroad"""

    env_args_map: dict = {
        "LIBERTY_FILE": "liberty-file",
        "LEF_TECH_FILE": "lef-tech-file",
        "LEF_FILE": "lef-file",
        "PIPE_STAGES": "pipe-stages",
        "RUN_OPENROAD": "run-openroad",
        "SYNTH_MODE": "synth-mode",
        "DELAY_TARGET_PS": "delay-target-ps",
        "SKIP_PASS2": "skip-pass2",
        # These are special, if the arg value is blank we'll auto fill it in from
        # parent classes Command/CommandDeisgn:
        "OUT_DIR": "work-dir",
        "TOP_MODULE": "top",
        "FILE_SUFFIX": "file-suffix",
    }

    sh_and_tcl_to_copy_to_work_dir: list = [
        "run_eda.sh",
        "syn_basic.tcl",
        "synth_pass1.tcl",
        "synth_pass2_targeted.tcl",
        "synth_pass2_full.tcl",
        "sta_analyze.tcl",
        "pnr_basic.tcl",
        "pnr_flow.tcl",
        "pdk_config.tcl",
        "ppa_metrics.tcl",
        "sta.tcl",
        "process_timing_metrics.py",
        "clock_ident.py",
    ]

    run_eda_sh__leaf = "run_eda.sh"
    run_eda_sh__fpath = str(
        (Path(os.path.dirname(__file__)) / "sh" / run_eda_sh__leaf).resolve()
    )

    def __init__(self, config: dict):
        CommandPnr.__init__(self, config=config)
        ToolOpenRoad.__init__(self, config=self.config)

        # Note: most of these have to work with the --docker-image, but these are easily overriden
        # with a .f file. These are also sent to the docker container via a generated env file, so
        # we also override these here if they are in your ENV.
        self.args.update({
            "docker-image": self.DEFAULT_DOCKER_IMAGE,
            "docker-env-keys": list(self.env_args_map.keys()),
            "liberty-file": "",
            "lef-tech-file": "",
            "lef-file": "",
            "pipe-stages": 1,
            "run-openroad": 1,
            "synth-mode": "twopass",
            "delay-target-ps": "",
            "skip-pass2": 0,
            "adaptive-clock": 0,
            "file-suffix": "",
            "clock-name": "",
            "clock-ns": "",
            "idelay-ns": "2",
            "odelay-ns": "2",
            "sta-sdc-file": "",
            "platform": "sky130hs",
            "stop-stage": "route",
            "place-density": "",
            "core-utilization": 65,
            "core-aspect-ratio": "1.0",
            "setup-slack-margin": 0,
            "hold-slack-margin": 0,
            "cts-buf-list": "",
            "max-wire-length": "",
            "run-eda-script": self.run_eda_sh__fpath,
        })

        self.args_help.update({
            "docker-image": "Docker image to use, required",
            "docker-env-keys": "environment variables that are passed to docker run command",
            "liberty-file": (
                "FILEPATH(s) to .lib (or .lib.gz) within the docker-image."
                " Space-delimited for multi-lib PDKs (e.g. ASAP7)."
                " If unset, auto-resolved from --platform."
            ),
            "lef-tech-file": (
                "FILEPATH to lef tech (tlef) file within the docker-image."
                " If unset, auto-resolved from --platform."
            ),
            "lef-file": (
                "FILEPATH(s) to lef file(s) within docker-image."
                " Space-delimited for multi-lef PDKs."
                " If unset, auto-resolved from --platform."
            ),
            "synth-mode": (
                "Synthesis mode: 'twopass' (default) for two-pass timing-driven synthesis,"
                " or 'basic' for the original single-pass flow."
            ),
            "delay-target-ps": (
                "ABC delay target in picoseconds for Pass 2 gate sizing."
                " If unset, auto-calculated as 85%% of clock period."
            ),
            "skip-pass2": (
                "Set to 1 to skip Pass 2 even if timing is violated after Pass 1."
                " Useful for quick area-only runs with the two-pass flow."
            ),
            "adaptive-clock": (
                "Set to 1 to enable adaptive clock period. When no --clock-ns is"
                " specified: Pass 1 uses a relaxed 10ns period, then tightens to"
                " 95%% of worst path delay after STA. Without this flag, unspecified"
                " clock-ns defaults to a PDK-appropriate period for a standard"
                " datapath (~15 logic levels). Has no effect when --clock-ns is set."
            ),
            "file-suffix": (
                "FILE_SUFFIX passed to docker run command, it is appeneded to the TOP_MODULE,"
                " if unset uses value from --seed."
            ),
            "platform": (
                "PDK platform identifier. Supported: sky130hs (default), asap7,"
                " sky130hd, nangate45, gf180, ihp-sg13g2. Controls floorplan site,"
                " routing layers, CTS buffers, PDN, tap cells, wire RC."
            ),
            "stop-stage": (
                "How far the PnR flow should progress: 'synth' (synthesis only),"
                " 'place' (through placement+timing repair), 'cts' (through clock tree),"
                " 'route' (full flow including detailed routing, default)."
            ),
            "place-density": (
                "Placement density override (0.0-1.0). If unset, uses the PDK default."
            ),
            "core-utilization": "Floorplan utilization target in percent (default: 65).",
            "core-aspect-ratio": "Floorplan aspect ratio (default: 1.0).",
            "setup-slack-margin": (
                "Setup timing repair margin in picoseconds (default: 0)."
            ),
            "hold-slack-margin": (
                "Hold timing repair margin in picoseconds (default: 0)."
            ),
            "cts-buf-list": (
                "CTS buffer cell list (space-separated). If unset, OpenROAD auto-detects."
            ),
            "max-wire-length": (
                "Max wire length in µm for repair_design. If unset, uses tool default."
            ),
            "run-eda-script": (
                "The default will use run_eda.sh from the pypackage opencos/tools/openroad/sh/,"
                " however you are free to bring-your-own script. If you have a local ./run_eda.sh"
                " that will be used instead of the pypackage one."
                ' This script is what is run via "docker run", volume mapped, in the container.'
            ),
        })

        self._update_args_with_current_env()

    def do_it(self) -> None:
        self.set_tool_defines()

        if not self.args["file-suffix"]:
            self.args["file-suffix"] = str(self.args.get("seed", "UNKNOWN"))

        self.write_eda_config_and_args()
        self._gen_docker_run_shell_command()

    def _set_run_eda_sh_script(self) -> None:
        """
        Finds string abspath of the args --run-eda-script=FILE or default value
        to use the one from the installed python package, and sets in
        self.run_eda_sh__fpath, self.run_eda_sh__leaf
        """
        try_file = Path(self.args["run-eda-script"])
        if try_file.is_file():
            self.run_eda_sh__fpath = str(try_file.resolve())
            self.run_eda_sh__leaf = str(try_file.name)
            return

        if try_file.parent == Path("."):
            # User provided a relative local file, or one w/out path, we can search in the
            # the pypackage space:
            try_installed = Path(os.path.dirname(__file__)) / "sh" / try_file.name
            if try_installed.is_file():
                self.run_eda_sh__fpath = str(try_installed.resolve())
                self.run_eda_sh__leaf = str(try_installed.name)
                return

            self.error(
                f"Cannot find file --run-eda-script={str(try_file)}, nor in",
                f"the installed defaults: {str(Path(os.path.dirname(__file__)) / 'sh')}",
            )
            return

        self.error(f"Cannot find file --run-eda-script={str(try_file)}")
        return

    def _update_args_with_current_env(self) -> None:
        """Called during __init__, updates self.args default values based on current ENV"""
        for env_name, arg_name in self.env_args_map.items():
            if value := os.environ.get(env_name, ""):
                # Note os.environ is str types, so check on bool/int/str
                cur_args_val = self.args[arg_name]
                if isinstance(cur_args_val, list):
                    # Don't know the type, so replace with a [str] and hope for the best
                    self.args[arg_name] = [value]
                elif isinstance(cur_args_val, int):
                    self.args[arg_name] = int(value)
                elif isinstance(cur_args_val, bool):
                    try:
                        self.args[arg_name] = bool(int(value))
                    except:  # pylint: disable=bare-except
                        self.args[arg_name] = bool(value)
                else:
                    self.args[arg_name] = value

    def _copy_scripts_to_work_dir(self) -> None:
        """Copies scripts to work dir"""
        # I guess we'll write out run_eda_sh__fpath to the work-dir, so user can re-run all
        # this via shell
        for x in self.sh_and_tcl_to_copy_to_work_dir:
            subdir = "."
            if x.endswith(".py"):
                pass
            elif x.endswith(".tcl"):
                subdir = "tcl"
            elif x.endswith(".sh"):
                subdir = "sh"
            else:
                self.error(
                    "unknow extension in CommandPnrOpenRoad.sh_and_tcl_to_copy_to_work_dir:",
                    f"{x}",
                )

            _fpath = str(Path(os.path.dirname(__file__)) / subdir / x)

            if not os.path.exists(_fpath):
                self.error(f"Cannot find required file in pypackage dist: {_fpath}")
            else:
                shutil.copy(_fpath, os.path.join(self.args["work-dir"], x))

        # Also copy self.run_eda_sh__fpath:
        shutil.copy(
            self.run_eda_sh__fpath,
            os.path.join(self.args["work-dir"], self.run_eda_sh__leaf),
        )

    def _gen_docker_run_shell_command(self) -> None:
        """Generates a docker_run.sh file (in work-dir) that will be used by: docker run"""

        work_dir_abs = Path(self.args["work-dir"]).resolve()
        self.args["work-dir"] = str(work_dir_abs)

        env_dict = {}
        for key in self.args["docker-env-keys"]:
            env_dict[key] = None
            if key in self.env_args_map:
                env_dict[key] = self.args[self.env_args_map[key]]

        # PDK library paths (LIBERTY_FILE, LEF_TECH_FILE, LEF_FILE) should be
        # resolved by run_eda.sh's resolve_pdk_libs() based on PLATFORM, unless
        # the user explicitly specified them via CLI args (--liberty-file, etc.).
        # Remove them from env_dict so that stale host environment variables
        # (e.g. from a previous sky130hs run) don't override the
        # platform-appropriate paths resolved by run_eda.sh.
        _pdk_lib_keys = ("LIBERTY_FILE", "LEF_TECH_FILE", "LEF_FILE")
        for _pk in _pdk_lib_keys:
            _arg_name = self.env_args_map.get(_pk, "")
            if not _arg_name:
                continue
            _val = self.args.get(_arg_name, "")
            _env_val = os.environ.get(_pk, "")
            # Keep the value only if the user explicitly set it on the CLI.
            # Heuristic: if the current value is empty OR matches the host env
            # var exactly, it wasn't set on the CLI — it's either the default
            # or came from _update_args_with_current_env.
            if not _val or _val == _env_val:
                env_dict.pop(_pk, None)

        # Instead of passing env vars (like VERILOG_FILES, ...) for files, incdirs,
        # defines, parameters to a docker container, instead
        # we will craft a .f for yosys or slang-plugin, using eda.
        # This should work even if the tool is not installed.

        flist_sco = CommandFListSlangYosys(config=self.config)
        flist_sco.set_tool_defines()
        for x in ("incdirs", "files_v", "files_sv", "parameters"):
            setattr(flist_sco, x, getattr(self, x))
        # for defines, we have to update with self.defines instead of copy, otherwise
        # you'll overwrite the tool defines from eda_config:
        getattr(flist_sco, "defines", {}).update(self.defines)

        # NOTE: this will not handle blackboxing as you might expect in DEPS.yml,
        # I'm not sure what to do about that yet, unless we wanted to support a user
        # BYO eda synth output(s). For now, still doing this one-shot
        yosys_flist_fpath = os.path.join(work_dir_abs, "eda_flist.yosys.read_slang.f")
        flist_sco.write_flist_to_file(fpath=yosys_flist_fpath)
        env_dict["YOSYS_READ_SLANG_DOTF"] = "eda_flist.yosys.read_slang.f"

        # send CLOCK_NAME, CLOCK_NS (period), STA_SDC_FILE, synthesis control,
        # PnR control, and PDK vars to the docker container
        for argname in (
            "clock-name",
            "clock-ns",
            "idelay-ns",
            "odelay-ns",
            "sta-sdc-file",
            "synth-mode",
            "delay-target-ps",
            "skip-pass2",
            "adaptive-clock",
            "platform",
            "stop-stage",
            "place-density",
            "core-utilization",
            "core-aspect-ratio",
            "setup-slack-margin",
            "hold-slack-margin",
            "cts-buf-list",
            "max-wire-length",
        ):
            if x := self.args[argname]:
                env_dict[argname.upper().replace("-", "_")] = x

        self._set_run_eda_sh_script()

        docker_run_base_cmd = docker_checks.get_docker_run_command(
            image=self.args["docker-image"],
            docker_command="run",
            user="$(id -u):$(id -g)",
            env_dict=env_dict,
            ro_volumes=get_source_files_paths(
                cmd_des_obj=self, minimal_root_paths=True
            ),
            rw_volumes=[work_dir_abs],
            threads=self.args["threads"],
        )

        self._copy_scripts_to_work_dir()

        util.write_shell_command_file(
            dirpath=self.args["work-dir"],
            filename="docker_run_edawrap_pnr.sh",
            command_lists=[
                util.ShellCommandList(
                    # Note the cd {work_dir_abs} is in the docker container:
                    docker_run_base_cmd
                    + [
                        "-c",
                        f"'cd {work_dir_abs} && ./{self.run_eda_sh__leaf}'",
                    ],
                    tee_fpath="docker_run.run_eda.log",
                )
            ],
            line_breaks=False,
        )

        # Promiscously save all .json that is saved in our work-dir by any TCL or SH scripts:
        util.artifacts.add_extension(
            search_paths=self.args["work-dir"], file_extension='json', typ='json',
            description='JSON data from PNR'
        )

        self.exec(
            work_dir=self.args["work-dir"], command_list=["./docker_run_edawrap_pnr.sh"]
        )



'''opencos.utils.str_helpers -- Various str helpers for printing, indenting'''

import fnmatch
import os
import re
import shlex
import textwrap

VALID_TARGET_INFO_STR: str = (
    "should start with a . or underscore/letter, rest should be"
    " ., alpha-numeric, dashes, or underscores."
)

# Parameters, with hierarchy, can have a name like: /Path/to.label[6]/Name
PARAMETER_NAME_RSTR : str = r'[\w\.\/\[\]]+'

def is_valid_target_name(s: str) -> bool:
    '''Returns True if str starts with . or underscore/letter, rest alphanum, dash, dot,

    or underscores. We allow '.' otherwise deps_file.py posts warnings about badly named
    targets for files that are missing.'''
    if not s or not (s[0].isalpha() or s[0] == '_' or s[0] == '.'):
        return False
    return s.replace('_', '').replace('-', '').replace('.', '').isalnum()

def strip_all_quotes(s: str) -> str:
    '''Returns str with all ' and " removed'''
    return s.replace("'", '').replace('"', '')


def strip_outer_quotes(s: str) -> str:
    '''Returns str with outer pairs of (' or ") removed

    Note this is done safely removing only outmost pairs of single quotes, or double
    quotes. This is used on bare CLI args that may have outer quotes from shlex.quote(arg).
    '''
    ret = str(s)
    while (ret.startswith("'") and ret.endswith("'")) or \
          (ret.startswith('"') and ret.endswith('"')):
        ret = ret[1:-1]
    return ret


def string_or_space(text: str, whitespace: bool = False) -> str:
    '''Returns str of either spaces (len(text)) or returns text.'''
    if whitespace:
        return " " * len(text)
    return text


def sanitize_defines_for_sh(value: object, shlex_quote: bool = False) -> str:
    '''Attempts to make a str for +define+key[=value] safer for using as a shell arg

    Need to sanitize this for shell in case someone sends a +define+foo+1'b0,
    which needs to be escaped as +define+foo+1\'b0, otherwise bash or sh will
    think this is an unterminated string.

    Optionally can use shlex.quote('+define+key=value') via shlex_quote=True
    '''
    if isinstance(value, str):
        value = value.replace("'", "\\" + "'")
    if shlex_quote:
        value = shlex.quote(value)
    return value


def sprint_time(time_value: int) -> str:
    '''Return pretty str for time'''
    s = int(time_value)
    txt = ""
    do_all = False
    # days
    if s >= (24 * 60 * 60): # greater than 24h, we show days
        d = int(s / (24 *60 *60))
        txt += f"{d}d:"
        s -= d * 24 * 60 * 60
        do_all = True
    # hours
    if do_all or s >= (60 *60):
        d = int(s / (60 * 60))
        txt += f"{d:2}:"
        s -= d * 60 * 60
        do_all = True
    # minutes
    d = int(s / 60)
    txt += f"{d:02}:"
    s -= d * 60
    # seconds
    txt += f"{s:02}"
    return txt


def indent_wrap_long_text(
        text: str, width: int = 80, initial_indent: int = 0, indent: int = 4
) -> str:
    """Returns str, wraps text to a specified width and indents subsequent lines."""
    wrapped_lines = textwrap.wrap(
        text, width=width,
        initial_indent=' ' * initial_indent,
        subsequent_indent=' ' * indent
    )
    return '\n'.join(wrapped_lines)


def dep_str2list(value: object) -> list:
    '''Helper for a markup \n or space separated string to be returned as a list'''
    if value is None:
        return []
    if isinstance(value, str):
        return re.split('\n+| +', value) # convert \n separated to list, also split on spaces
    return value


def fnmatch_or_re(pattern: str, string: str) -> bool:
    '''Returns True if pattern/string matches in re.match or fnmatch'''
    matches = []
    # fnmatch check, aka: ./*test
    matches.append(
        bool(fnmatch.fnmatch(name=string, pat=pattern))
    )
    # regex check, aka: ./.*test
    try:
        matches.append(
            bool(re.match(pattern=pattern, string=string))
        )
    except: # pylint: disable=bare-except
        # could have been an illegal/unsupported regex, so don't match.
        pass
    return any(matches)


def match_glob_patterns(name: str, patterns: list) -> bool:
    """Check if name matches any of the given glob patterns (case-insensitive)."""
    return any(
        fnmatch.fnmatch(name.lower(), pattern.lower()) for pattern in patterns
    )


def get_terminal_columns():
    """
    Retrieves the number of columns (width) of the terminal window.

    Returns:
        int: The number of columns in the terminal, or a default value (e.g., 80)
             if the terminal size cannot be determined. Min value of 40 is returned.
    """
    try:
        size = os.get_terminal_size()
        return max(40, size.columns)
    except OSError:
        # Handle cases where the terminal size cannot be determined (e.g., not in a TTY)
        return 80  # Default to 80 columns

    return 80 # else default to 80.


def pretty_list_columns_manual(
        data: list, num_columns: int = 4, auto_columns: bool = True, _prev_recursion: int = 0
) -> list:
    """Returns list, from list of str, organized into columns, manually aligning them.

    prev_recursion is to let us know if we have added AND removed columns, to avoid situations
    where the recursion would otherwise never complete (because the 2D matrix is jumbled every
    recursion attempt.)
       0 initially,
       0x1 ([0] bit) set if we have subtracted a column,
       0x2 ([1] bit) set if we have added a column.
       If 0x3 (binary 11) then we have added and subtracted, please stop the recursion.
    """

    ret_lines = []
    if not data:
        return ret_lines

    _spacing = 2

    # Calculate maximum width for each column
    max_lengths = [0] * num_columns
    max_item_len = 0
    for i, item in enumerate(data):
        col_index = i % num_columns
        max_lengths[col_index] = max(max_lengths[col_index], len(item))
        max_item_len = max(max_item_len, len(item))

    if _prev_recursion != 0x3 and auto_columns and num_columns > 1:
        window_cols = get_terminal_columns()
        max_line_len = 0
        for x in max_lengths:
            max_line_len += x + _spacing
        if max_line_len >= window_cols:
            # subtract a column (already >= 2):
            # TODO(drew): bug here on recursion, we can end in a death spiral on the return, and the
            # one in the below if-stmt. DEBUG
            ret_lines.extend(
                pretty_list_columns_manual(data=data, num_columns=num_columns-1, auto_columns=True,
                                           _prev_recursion=_prev_recursion | 0x1)
            )
            return ret_lines
        if (max_line_len + max_item_len + _spacing) < window_cols:
            # add 1 more column if we're guaranteed to have room.
            ret_lines.extend(
                pretty_list_columns_manual(data=data, num_columns=num_columns+1, auto_columns=True,
                                           _prev_recursion=_prev_recursion | 0x2)
            )
            return ret_lines
        # else continue

    # Print data in columns
    ret_lines.append('')
    for i, item in enumerate(data):
        col_index = i % num_columns
        ret_lines[-1] += str(item).ljust(max_lengths[col_index] + _spacing)
        if col_index == num_columns - 1 or i == len(data) - 1:
            ret_lines.append('')

    return ret_lines

def pretty_2dlist_columns(
        data: list, padding: int = 2, return_as_2d_list: bool = False,
        header_row_centered: bool = False
) -> list:
    '''Given a list of list-of-str, return a 1d list with values aligned in columns

    If return_as_2d_list=True, does not join the padded row items.
    '''

    if not data:
        return []

    ret_lines = []
    # include padding in the col widths:
    col_widths = [max(len(str(value)) for value in col) + padding for col in zip(*data)]

    if return_as_2d_list:
        for rownum, row in enumerate(data):
            if header_row_centered and rownum == 0:
                ret_lines.append([
                    f'{item:^{col_widths[col_idx]}}' for col_idx,item in enumerate(row)
                ])
            else:
                ret_lines.append([
                    f'{item:<{col_widths[col_idx]}}' for col_idx,item in enumerate(row)
                ])
        return ret_lines

    # else return as 1D list with everything already formatted.
    header_fmt = ''.join(f"{{:^{w}}}" for w in col_widths)
    format_str = ''.join(f"{{:<{w}}}" for w in col_widths)
    for rownum, row in enumerate(data):
        if header_row_centered and rownum == 0:
            ret_lines.append(header_fmt.format(*row))
        else:
            ret_lines.append(format_str.format(*row))

    return ret_lines

def print_columns_manual(data: list, num_columns: int = 4, auto_columns: bool = True) -> None:
    """Prints a list of strings in columns, manually aligning them."""

    lines = pretty_list_columns_manual(
        data=data, num_columns=num_columns, auto_columns=auto_columns
    )
    print('\n'.join(lines))

def strip_ansi_color(text: str) -> str:
    '''Strip ANSI color characters from str'''
    ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?;]*[ -/]*[@-~])')
    return ansi_escape.sub('', text)

def get_shorter_path_str_rel_vs_abs(rel_path: str) -> str:
    '''Returns the shorter of relative path (input arg) vs abspath (converted)'''
    apath = os.path.abspath(rel_path)
    if len(apath) < len(rel_path):
        return apath
    return rel_path


def flatten_dict_to_str(data, parent_key='', sep='.', item_sep=', ') -> str:
    '''Attemps to flatten a dict into: key1.subkey2 = value, ...'''
    items = []
    for k, v in data.items():
        # Create the new key (e.g., "tool_errors.verilator")
        new_key = f"{parent_key}{sep}{k}" if parent_key else k

        if isinstance(v, dict):
            if v:
                # remove empty dict entries
                # Recurse into nested dictionaries
                items.append(flatten_dict_to_str(v, new_key, sep=sep, item_sep=item_sep))
        elif isinstance(v, list):
            for list_index, list_item in enumerate(v):
                if list_item:
                    items.append(flatten_dict_to_str(
                        list_item, f'{new_key}[{list_index}]', sep=sep, item_sep=item_sep
                    ))

        else:
            # Add the final key-value pair
            items.append(f"{new_key}: {v}")

    return item_sep.join(items)



'''opencos.commands.show_tools - command handler for: eda show_tools [args]

Note this command is handled differently than others (such as CommandSim),
it is generally run as simply

    > eda show_tools

uses no tools and will print a pretty list of tools installed (similar to --help)
'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

import json
import os

from opencos import util, eda_tool_helper
from opencos.eda_base import Command


class CommandShowTools(Command):
    '''command handler for: eda targets'''

    command_name = 'targets'

    def __init__(self, config: dict):
        super().__init__(config)
        # We don't have any args, clear them, and add a few we'll support
        # --quiet handled by opencos.util
        # --json-to-stdout handled by us
        self.args = {
            'json-to-stdout': False,
            'json-out': '',
        }
        self.args_help = {
            'json-to-stdout': 'Prints tool information as JSON to STDOUT',
            'json-out': 'Optional JSON filename to save output',
        }
        self.config = config
        self.status = 0

    def process_tokens( # pylint: disable=unused-argument
        self, tokens: list, process_all: bool = True,
        pwd: str = os.getcwd()
    ) -> list:
        '''This is effectively our 'run' method, entrypoint from opencos.eda.main'''

        # disable artifacts, logging is disabled by eda.py adding 'deps-help' to:
        #   util.global_log.default_log_disable_with_args
        util.artifacts.enabled = False

        # run with process_all=False, we don't care about unparsed args
        super().process_tokens(tokens=tokens, process_all=False, pwd=pwd)

        if any(self.args[a] for a in ('json-to-stdout', 'json-out')):
            self._print_tools_as_json()

        return []

    def _print_tools_as_json(self) -> None:
        '''Prints a JSON string to STDOUT for tool, path, version'''

        info = eda_tool_helper.get_handler_info_with_versions(
            config=self.config, include_commands=True, sort=True, all_details=True
        )
        if self.args['json-to-stdout']:
            print(json.dumps(info, indent=4))
        if filepath := self.args['json-out']:
            try:
                with open(filepath, 'w', encoding='utf-8') as f:
                    f.write(json.dumps(info, indent=4))
                util.info(f'Wrote tool information --json-out JSON file: {filepath}')
            except PermissionError:
                self.error(f'PermissionError trying to write --json-out JSON file: {filepath}')
            except IOError as e:
                self.error(f'IOError while trying to write --json-out JSON file: {filepath}, {e}')



'''opencos.commands.sim - Base class command handler for: eda sim ...

Intended to be overriden by Tool based classes (such as CommandSimVivado, etc)

Note that CommandSim is also a base class for opencos.commands.elab.CommandElab.'''

# Note - similar code waiver, tricky to eliminate it with inheritance when
# calling reusable methods.
# pylint: disable=R0801

# pylint: disable=too-many-arguments

import os
from pathlib import Path
import re
import shlex


from opencos import util
from opencos.eda_base import CommandDesign, Tool
from opencos.utils import status_constants

from opencos.utils.str_helpers import strip_outer_quotes

THISPATH = os.path.dirname(__file__)
WAVES_PKG_SV_FNAME = str((Path(THISPATH) / '..' / '_waves_pkg.sv').resolve())

# Unique checks for UVM Report
UVM_ERROR_REPORTING_RE = r"^.* UVM_ERROR\s*:\s*0*([1-9]\d*)\s*$"
UVM_FATAL_REPORTING_RE = r"^.* UVM_FATAL\s*:\s*0*([1-9]\d*)\s*$"
UVM_WARNING_REPORTING_RE = r"^.* UVM_WARNING\s*:\s*0*([1-9]\d*)\s*$"

# Check for UVM_ERROR /some/path, or UVM_ERROR: /some/path
# Avoids the reporting section: UVM_ERROR : 1
UVM_ERROR = r"UVM_ERROR\s*\:?\s*[^\s\:0-9]"
UVM_FATAL = r"UVM_FATAL\s*\:?\s*[^\s\:0-9]"
UVM_WARNING = r"UVM_WARNING\s\:?\s*[^\s\:0-9]"

def parameters_dict_get_command_list(
        params: dict, arg_prefix: str = '-G',
        hier_delimiter: str = '.',
        top_hier_str: str = '',
        for_flist: bool = False
) -> list:
    '''Given dict of parameters, returns a command list.

    Some simulators will set ALL parameters with Name=Value at all levels of hierarchy,
    so to avoid that and only set top level parameters, but still allow package or
    hierarchy parameters to be set, you can use:

    hier_delimiter='/'
    top_hier_str=f'/{TOP_MODULE_NAME}/'

    Which will, if a parameter Name does not contain the heir delimiter, the Name
    will be prepended with top_hier_str.
    '''

    ret_list = []
    if ' ' in arg_prefix:
        arg_list_prefix = arg_prefix.split()
        arg_str_prefix = ''
    else:
        arg_list_prefix = []
        arg_str_prefix = arg_prefix

    for k,v in params.items():
        if not isinstance(v, (int, str)):
            util.warning(f'parameter {k} has value: {v}, parameters must be int/string types')

        if top_hier_str and hier_delimiter not in k:
            k = f'{top_hier_str}{k}'

        ret_list.extend(arg_list_prefix)
        if isinstance(v, int):
            ret_list.append(f'{arg_str_prefix}{k}={v}')
        else: # string
            # Note if for_flist=True, then we have to take extra care on what gets
            # double-quotes " and \".
            v = strip_outer_quotes(v.strip('\n'))
            v = '"' + v + '"'

            if for_flist:
                # do not wrap with " unless there are spaces, if there are spaces
                # wrap the entire thing with single quotes (bash)
                if ' ' in v:
                    ret_list.append(shlex.quote(f'{arg_str_prefix}{k}={v}'))
                else:
                    ret_list.append(f'{arg_str_prefix}{k}={v}')
            else:
                # This is for a tool, so simply guard single quotes:
                v = v.replace("'", "\\" + "'")
                ret_list.append(f'{arg_str_prefix}{k}={v}')
    return ret_list


class CommandSim(CommandDesign): # pylint: disable=too-many-public-methods
    '''Base class command handler for: eda sim ...

    The common entrypoint is from eda.py (opencos.eda):
    ***--> process_tokens()
             --> create work-dir, fail early
             --> run commands from DEPS
             --> run_argparser_on_list() -- populate self.args
          ***--> run() / do_it()
                 --> prepare_compile()
                       --> get_*_command_lists(), or similar
                 --> write out eda config to work-dir for debug
                 --> handle "export" args if present
                 --> compile(), elaborate(), simulate()
             --> run_post_tool_dep_commands()
             --> report pass/fail
             --> return unparsed args back to eda.py.

    Note that this flow is also often used by CommandElab (eda elab ..., or eda lint ...)
    And follows a similar pattern, except usually falls short of elaborate() or simulate()
    steps in self.run()
    '''

    CHECK_REQUIRES = [Tool] # Used by check_command_handler_cls()
    error_on_no_files_or_targets = True
    error_on_missing_top = True
    tool_config = {} # Children with Tool parent classes will set on Tool constructor.
    library_map_supported = False
    has_pre_compile_slang_sh: str = '' # if set, save in all.sh before compile.sh

    command_name = 'sim'

    def __init__(self, config: dict):
        CommandDesign.__init__(self, config=config)
        self.args.update({
            "pre-sim-tcl": [],
            'compile-args': [],
            'compile-waivers': [],
            'elab-args': [],
            'sim-args': [],
            'sim-plusargs': [], # lists are handled by 'set_arg(k,v)' so they append.
            'sim-library': [],
            'sim-waivers': [],
            'coverage': False,
            'waves': False,
            'waves-start': 0,
            'pass-pattern': "",
            'optimize': False,
            'log-bad-strings': ['ERROR: ', 'FATAL: ', 'Error: ', 'Fatal: '],
            'log-must-strings': [],
            'log-warning-strings': ['WARNING: ', 'Warning: '],
            'uvm': False,
            'uvm-version': '1.2',
            # verilate-args: list of args you can only pass to Verilator,
            # not used by other simulators, so these can go in DEPS files for custom things
            # like -CFLAGS -O0, etc.
            'verilate-args': [],
            'ext-defines-sv-fname': '',
            'license-queue': False,
            'license-queue-value': '30',
            'library-map': [],
            'add-top-library': [],
            'docker-run': False,
            'docker-image': ''
        })
        self.args_help.update({
            'pre-sim-tcl': (
                'Tool dependent, TCL file to run prior to simulation (if tool supports TCL)'
            ),
            'compile-args': 'args added to sim/elab "compile" step',
            'compile-waivers': (
                'simulation tool depenent waivers by name, applied at compilation step'
            ),
            'coverage': 'attempt to run coverage steps on the compile/elab/simulation',
            'elab-args':    'args added to sim/elab "elaboration" step, if required by tool',
            'log-bad-strings': 'strings that if present in the log will fail the simulation',
            'log-must-strings': ('strings that are required by the log to not-fail the simulation.'
                                 ' Some tools use these at only certain phases'
                                 ' (compile/elab/sim).'),
            'log-warning-strings': ('strings that if present will be counted in the total tool'
                                    ' warnings at end of simulation'),
            'pass-pattern': ('Additional string required to pass a simulation, appends to'
                             ' log-must-strings'),
            'sim-args':      'args added to final "simulation" step',
            'sim-plusargs':  ('"simulation" step run-time args passed to tool, these can also'
                              ' be set using --sim-plusargs=name[=value], or simply +name[=value]'),
            'sim-library': 'named libraries added to simulation, tool dependent',
            'sim-waivers': (
                'simulation tool depenent waivers by name, applied at "simulation" step'
            ),
            'stop-before-compile': ('Create work-dir sh scripts for compile/elab/simulate, but do'
                                    ' not run them.'),
            'stop-after-compile': 'Create work-dir sh scripts, but only run the compile step.',
            'stop-after-elaborate': ('Create work-dir sh scripts, but run compile+elab, skip'
                                     ' simulation step.'),
            'top': 'Name of topmost Verilog/SystemVerilog module, or VHDL entity',
            'verilate-args': ('args added to "compile" step in Verilator simulation'
                              ' (for --tool=verilator) these are identical to --compile-args'),
            'waves': 'Include waveforms, if possible for tool',
            'waves-start': 'Starting time of waveform capture, if possible for tool',
            'test-mode': ('stops the command early without executing, if --gui is present will'
                          ' instead test without spawning gui'),
            'uvm': (
                'Attempts to support UVM (aka 1.2 if --uvm-version=1.2) for the'
                ' the simulation tool. May add libraries for uvm, tool dependent.'
            ),
            'uvm-version': (
                'Used if --uvm is set, for example --uvm-version=1.2'
            ),
            'ext-defines-sv-fname': (
                'If set to empty str, --ext-defines-sv-fname="", applies defines to tool via args.'
                ' If set to non-emtpy str as a .sv file name, such as'
                ' --ext-defines-sv-fname=_ext_defines.sv, saves defines to a file name in the'
                ' work-dif, and adds this file to the SV/Verilog file list ahead of other files.'
            ),
            'license-queue': (
                'Tool dependent, if supported help will show more details:'
                ' eda COMMAND --help --license-queue'
            ),
            'license-queue-value': (
                'Tool dependent, value used internally if --license-queue set'
            ),
            'library-map': (
                'list arg (can set multiple) with values LibName:Path, only supported'
                ' by certain simulation tools.'
                ' The Path search is relative to the tool exe, one level above the tool exe,'
                ' or can be absolute path.'
            ),
            'add-top-library': (
                'Add additional top levels, for example if you added the "unisim" library'
                ' you may need to --add-top-library=unisim.glbl for it to work with Xilinx'
                ' based projects that require "glbl" to be present at simulation $root scope.'
            ),
            'docker-run': (
                'Run the simulation using "docker run" with image specified by'
                ' --docker-image=NAME. Assumes the entrypoint can be /bin/bash and your --tool'
                ' executable is installed in PATH. NOTE: docker-in-docker is not supported yet,'
                ' if you are already in docker this option is not allowed and will fail checks.'
            ),
            'docker-image': (
                'Used with arg --docker. IMAGE used by "docker run" for the sim, lint, or'
                ' elab eda command.'
            )

        })

        self.args_kwargs.update({
            'uvm-version': { 'choices': ['1.2'] },
        })

        # only used if self.library_map_supported=True:
        self.library_map = {} # lib-name: path-to.lib dict from self.args['library-map']


    def process_parameters_get_list(
            self, arg_prefix: str = '-G', **kwargs
    ) -> list:
        '''Returns list (suitable command list for shell or for tool) from self.parameters

        Some simulators will set ALL parameters with Name=Value at all levels of hierarchy,
        so to avoid that and only set top level parameters, but still allow package or
        hierarchy parameters to be set, you can use args (in kwargs):

        hier_delimiter='/'
        top_hier_str=f'/{TOP_MODULE_NAME}/'

        Which will, if a parameter Name does not contain the heir delimiter, the Name
        will be prepended with top_hier_str.
        '''
        return parameters_dict_get_command_list(
            params=self.parameters, arg_prefix=arg_prefix, **kwargs
        )

    def process_plusarg(self, plusarg: str, pwd: str = os.getcwd()) -> None:
        '''Override for CommandDesign.process_plusarg(..)'''
        maybe_plusarg = CommandDesign.process_plusarg(self, plusarg, pwd)
        # Support for self.args['unprocessed-plusargs'] --> self.args['sim-plusargs']:
        if maybe_plusarg and \
           maybe_plusarg in self.args['unprocessed-plusargs'] and \
           maybe_plusarg not in self.args['sim-plusargs']:
            self.args['sim-plusargs'].append(maybe_plusarg)
            self.args['unprocessed-plusargs'].remove(maybe_plusarg)
            util.debug(f'For parent "sim" command (CommandSim), moved plusarg: {maybe_plusarg}',
                       'to sim-plusargs (from unprocessed-plusargs)')

    def process_tokens(self, tokens: list, process_all: bool = True,
                       pwd: str = os.getcwd()) -> list:
        self.defines['SIMULATION'] = None
        unparsed = CommandDesign.process_tokens(
            self, tokens=tokens, process_all=process_all, pwd=pwd
        )

        if self.stop_process_tokens_before_do_it():
            return unparsed

        # add defines for this job type
        if self.args['lint'] or self.args['stop-after-elaborate']:
            self.args['lint'] = True
            self.args['stop-after-elaborate'] = True
        if self.args['top']:
            # create our work dir
            self.create_work_dir()
            self.run_dep_commands()
            self.do_it()
            self.run_post_tool_dep_commands()
            self.report_pass_fail()
        return unparsed


    def set_tool_config_from_config(self) -> None:
        '''Sets self.tool_config (from original --config-yml=YAML|Default) and overrides
        self.defines, and self.args log-must-strings and log-bad-strings.'''
        tool = self.args.get('tool', '') # get from Command's self.args['tool']
        if tool:
            self.tool_config = self.config.get('tools', {}).get(tool, {})
            self.override_log_strings_from_tool_config()
            self.defines.update(self.tool_config.get('defines', {}))
            util.debug(f'set_tool_config_from_config: {tool=}')

    def update_tool_config(self):
        self.override_log_strings_from_tool_config()


    def override_log_strings_from_tool_config(self) -> None:
        '''Returns None, overrides memvers of self.args based on our tool_config

        (tool_config from eda.py --config-yml=YAML-CONFIG-FILE)
        '''
        if not getattr(self, 'tool_config', None):
            return

        # Collect (overwrite CommandSim) the bad and must strings, if present,
        # from our config.tools.verilator:
        for tool_config_key in ('log-bad-strings', 'log-must-strings', 'log-warning-strings'):
            if len(self.tool_config.get(tool_config_key, [])) > 0:
                self.args[tool_config_key] = self.tool_config.get(tool_config_key, [])


    # Methods that derived classes may override:

    def run_commands_check_logs( # pylint: disable=dangerous-default-value,too-many-locals
            self, commands: list , check_logs: bool = True, log_filename=None,
            bad_strings: list = [],
            must_strings: list = [],
            use_bad_strings: bool = True, use_must_strings: bool = True,
            sim_step: str = ''
    ) -> None:
        '''Returns None, runs all commands (each element is a list) and checks logs

        for bad-strings and must-strings (args or class member vars).

        Will add any bad_strings or tool log-warning-strings to self, so tool
        related warning/error counts can be reported later.

        sim_step can be one of: compile, elaborate, simulate, or blank
        '''

        for obj in commands:

            assert isinstance(obj, list), \
                (f'{self.target=} command {obj=} is not a list or util.ShellCommandList,'
                 ' not going to run it.')

            clist = list(obj).copy()
            tee_fpath = getattr(obj, 'tee_fpath', None)
            work_dir = getattr(obj, 'work_dir', None)
            if not work_dir:
                work_dir = self.args['work-dir']

            util.debug(f'run_commands_check_logs: {clist=}, {tee_fpath=}, {log_filename=}')

            log_fname = None
            if tee_fpath:
                log_fname = tee_fpath
            if log_filename:
                log_fname = log_filename

            # track the retcode because we run stop_on_error=False (to count warnings/errors)
            _, stdout, retcode = self.exec(
                work_dir=work_dir, command_list=clist, tee_fpath=tee_fpath,
                stop_on_error=False,
            )

            if check_logs and not log_fname and bad_strings:
                util.warning(
                    f'{self.get_info_job_name()}: check_logs=True but no log file set',
                    f'(run --debug for more info), for command: {" ".join(clist)}'
                )

            if check_logs and log_fname:
                # Note this call will check on stdout if not GUI, not opening the log_fname,
                # but if this is GUI we normally lose stdout and have to open the log.
                if self.args.get('gui', False):
                    file_contents_str = ''
                else:
                    file_contents_str = stdout

                self.check_logs_for_errors(
                    sim_retcode=retcode,
                    filename=os.path.join(work_dir, log_fname),
                    file_contents_str=file_contents_str,
                    bad_strings=bad_strings, must_strings=must_strings,
                    use_bad_strings=use_bad_strings, use_must_strings=use_must_strings,
                    sim_step=sim_step
                )
            if log_fname:
                self.artifacts_add(
                    name=os.path.join(work_dir, log_fname),
                    typ='text', description='Simulator stdout/stderr log file'
                )
            if retcode:
                self.error(
                    f"{self.get_info_job_name()}: exec returned with error code: {retcode}"
                )

    def do_export(self, **kwargs) -> None:
        '''CommandSim helper for handling args --export*, override from Command'''
        super().do_export(
            deps_file_args=[
                '--compile-args',
                '--elab-args',
                '--sim-',
                '--coverage',
                '--waves',
                '--pass-pattern',
                '--optimize',
                '--stop-',
                '--lint-',
                '--verilat',
            ],
            support_export_run=True
        )


    def do_it(self) -> None:
        self.prepare_compile()
        self.write_eda_config_and_args()

        if self.is_export_enabled():
            # If we're exporting the target, we do NOT run the test here
            # (do_export() may run the test in a separate process and
            # from the out_dir if --export-run was set)
            self.do_export()
            return

        self.compile()
        self.elaborate()
        self.simulate()

    # Methods that derived classes may override:

    def prepare_compile(self) -> None:
        '''Derived classes may override if they want to create .sh scripts in the work-dir.

        Common use-case is to call:
            self.set_tool_defines()
            cmds0 = self.get_compile_command_lists()
            cmds1 = self.get_elaborate_command_lists()
            cmds2 = self.get_simulate_command_lists()
            cmds3 = self.get_post_simulate_command_lists()
        '''
        self.update_library_map()


    def check_lines_for_uvm_reporting(self, log_lines: list, log_fname: str) -> None:
        '''Scrapes lines looking for UVM reporting, adds to tool warning/errors'''
        if not self.args['uvm']:
            return

        # We'd like to check for UVM reporting, which we'll handle outside of _bad_strings
        # or _warning_strings b/c we have to fish out the count use regex:
        for lineno, line in enumerate(log_lines):
            me = re.match(UVM_ERROR_REPORTING_RE, line)
            mf = re.match(UVM_FATAL_REPORTING_RE, line)
            mw = re.match(UVM_WARNING_REPORTING_RE, line)
            if me:
                _count = me.group(1)
                self.incr_tool_error(int(_count))
                self.error(
                    f'log {log_fname}:{lineno+1} observed UVM reporting errors: {_count}',
                    do_exit=False, error_code=status_constants.EDA_SIM_UVM_ERROR_OR_FATAL
                )
            if mf:
                _count = mf.group(1)
                self.incr_tool_error(int(_count))
                util.error(
                    f'log {log_fname}:{lineno} observed UVM reporting fatals: {_count}',
                    do_exit=False, error_code=status_constants.EDA_SIM_UVM_ERROR_OR_FATAL
                )
            if mw:
                _count = mw.group(1)
                self.incr_tool_warning(int(_count))
                util.warning(
                    f'log {log_fname}:{lineno} observed UVM reporting warnings: {_count}'
                )

    def _do_log_scrape_lines(self, lines: list, must_strings: list, bad_strings: list,
                             log_fname: str, sim_step: str) -> None:
        '''Performs the actual Must-have checks and Bad strings checks, on list "lines"'''

        hit_must_string_dict = dict.fromkeys(must_strings)

        # Pre-compile some regex for better performance
        combined_regex_bad_strings = None
        combined_regex_must_strings = None
        if bad_strings:
            combined_regex_bad_strings = re.compile('|'.join(f'({p})' for p in bad_strings))
        if must_strings:
            combined_regex_must_strings = re.compile('|'.join(f'({p})' for p in must_strings))

        if lines:
            for lineno, line in enumerate(lines):

                if combined_regex_must_strings and combined_regex_must_strings.search(line):
                    for _must_str in must_strings:
                        if re.search(_must_str, line): # found it
                            hit_must_string_dict[_must_str] = True

                if combined_regex_bad_strings and combined_regex_bad_strings.search(line):
                    _bad_str = ''
                    for _bad_str in bad_strings:
                        if re.search(_bad_str, line):
                            break # found the offending one
                    self.error(
                        f"{sim_step}log {log_fname}:{lineno+1} contains: {_bad_str}",
                        error_code=status_constants.EDA_SIM_LOG_HAS_BAD_STRING
                    )

        if any(x is None for x in hit_must_string_dict.values()):
            self.error(
                f"{sim_step}Didn't get all passing patterns in log {log_fname}: {must_strings},",
                f" {hit_must_string_dict}",
                error_code=status_constants.EDA_SIM_LOG_MISSING_MUST_STRING
            )


    # pylint: disable=too-many-locals,too-many-branches,too-many-statements
    def check_logs_for_errors( # pylint: disable=dangerous-default-value
            self,
            sim_retcode: int = 0,
            filename: str = '',
            file_contents_str: str = '',
            bad_strings: list = [], must_strings: list = [],
            warning_strings: list = [],
            use_bad_strings: bool = True, use_must_strings: bool = True,
            sim_step: str = ''
    ) -> None:
        '''Returns None, checks logs using args bad_strings, must_strings,

        and internals self.args["log-[bad|must]-strings"] (lists).

        file_contents_str will take precedence over opening filename, but filename is still
        used in messaging.

        sim_step can be one of: compile, elaborate, simulate, or blank
        '''

        _bad_strings = bad_strings
        _must_strings = must_strings
        # append, if not they would 'replace' the args values:
        if use_bad_strings:
            _bad_strings = bad_strings + self.args.get('log-bad-strings', [])
        if use_must_strings:
            _must_strings = must_strings + self.args.get('log-must-strings', [])

        _warning_strings = warning_strings + self.args.get('log-warning-strings', [])

        if self.args['pass-pattern'] != "":
            _must_strings.append(self.args['pass-pattern'])

        if self.args['uvm']:
            _bad_strings.extend([
                UVM_ERROR, UVM_FATAL
            ])
            _warning_strings.append(UVM_WARNING)

        if sim_step:
            sim_step = f'({sim_step}) '

        if len(_bad_strings) == 0 and len(_must_strings) == 0:
            return

        lines = []

        log_fpath = ''
        if os.path.isfile(filename):
            log_fpath = filename

        if file_contents_str:
            lines = file_contents_str.split('\n')
            log_fname = log_fpath + '(STDOUT)'
            util.debug(f'{sim_step}Checking log for errors: {log_fpath=} but checking from',
                       'STDOUT string')
        elif log_fpath:
            log_fname = log_fpath
            util.info(f'{sim_step}Checking log filename: {log_fpath}')
            with open(log_fpath, 'r', encoding='utf-8') as f:
                lines = f.read().splitlines()
        else:
            log_fname = filename
            self.error(f'sim.check_logs_for_errors: {sim_step}{log_fname=} is not a file or does',
                       'not exist, and no file_contents_str exists to check')

        self.update_tool_warn_err_counts_from_log_lines(
            log_lines=lines, bad_strings=_bad_strings, warning_strings=_warning_strings
        )

        self.check_lines_for_uvm_reporting(log_lines=lines, log_fname=log_fname)

        func = getattr(self, 'report_tool_warn_error_counts', None)
        if func and isinstance(self, Tool) and callable(func):
            self.report_tool_warn_error_counts(clear_counts=True) # pylint: disable=no-member

        if sim_retcode > 0:
            # We have to update artifacts first, have the caller set the error.
            # Skip the checking for bad strings or must strings, because we've already
            # failed, but we did the important counting of Error and Warning lines above,
            # and reported it.
            return

        self._do_log_scrape_lines(
            lines=lines, must_strings=_must_strings, bad_strings=_bad_strings,
            log_fname=log_fname, sim_step=sim_step
        )


    def write_sh_scripts_to_work_dir(
            self, compile_lists: list, elaborate_lists: list, simulate_lists: list,
            compile_line_breaks: bool = True,
            elaborate_line_breaks: bool = False,
            simulate_line_breaks: bool = False,
            simulate_sh_fname: str = 'simulate.sh',
            all_sh_fname: str = 'all.sh'
    ) -> None:
        '''Writes compile.sh, elaborate.sh, simulate.sh (if present), all.sh to work-dir

        Will include the pre_compile_dep_shell_commands.sh if those are present.
        compile_line_breaks defaults to True (one word per line w/ line breaks added)
        '''

        all_lists = [] # list - of - (command-list)
        if self.has_pre_compile_dep_shell_commands:
            all_lists.append(
                ['./pre_compile_dep_shell_commands.sh']
            )

        if self.has_pre_compile_slang_sh and \
           os.path.exists(os.path.join(self.args['work-dir'], self.has_pre_compile_slang_sh)):
            # Derived classes would have run Slang on their own if needed, but for repro
            # purposes we'll add it to all.sh:
            all_lists.append(
                [f'./{self.has_pre_compile_slang_sh}']
            )

        if compile_lists:
            util.write_shell_command_file(dirpath=self.args['work-dir'], filename='compile.sh',
                                          command_lists=compile_lists,
                                          line_breaks=compile_line_breaks)
            all_lists.append(['./compile.sh'])

        if elaborate_lists:
            util.write_shell_command_file(dirpath=self.args['work-dir'], filename='elaborate.sh',
                                          command_lists=elaborate_lists,
                                          line_breaks=elaborate_line_breaks)
            all_lists.append(['./elaborate.sh'])

        if simulate_lists:
            util.write_shell_command_file(dirpath=self.args['work-dir'], filename=simulate_sh_fname,
                                          command_lists=simulate_lists,
                                          line_breaks=simulate_line_breaks)
            all_lists.append(['./' + simulate_sh_fname])

        if self.has_post_tool_dep_shell_commands:
            all_lists.append(
                ['./post_tool_dep_shell_commands.sh']
            )

        util.write_shell_command_file(dirpath=self.args['work-dir'], filename=all_sh_fname,
                                      command_lists=all_lists)

        self.write_eda_config_and_args()



    # Methods that derived classes must override:

    def compile(self) -> None:
        '''compile() runs as part of the CommandSim.do_it() flow'''
        raise NotImplementedError

    def elaborate(self) -> None:
        '''elaborate() runs as part of the CommandSim.do_it() flow, after compile()'''
        raise NotImplementedError

    def simulate(self) -> None:
        '''simulate() runs as part of the CommandSim.do_it() flow, after elaborate()'''
        raise NotImplementedError

    def get_compile_command_lists(self, **kwargs) -> list:
        ''' Returns a list of lists (list of command lists).'''
        raise NotImplementedError

    def get_elaborate_command_lists(self, **kwargs) -> list:
        ''' Returns a list of lists (list of command lists).'''
        raise NotImplementedError

    def get_simulate_command_lists(self, **kwargs) -> list:
        ''' Returns a list of lists (list of command lists).'''
        raise NotImplementedError

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        ''' Returns a list of lists (list of command lists).'''
        raise NotImplementedError

    def add_waves_pkg_file(self) -> None:
        '''If --waves is present, and one of --waves-fst or --waves-vcd or --dump-vcd, and

        the user is missing any $dumpfile(), then adds a pre-written
        SystemVerilog package to their source code. Note that individual tools have
        to call this prior to their compile step, CommandSim does not run this method for
        you.
        '''
        if not self.args['waves']:
            return
        if not any(self.args.get(x, False) for x in ('waves-fst', 'waves-vcd', 'dump-vcd')):
            return
        found_dumpfile = False
        for fname in self.files_v + self.files_sv:
            if found_dumpfile:
                break
            with open(fname, encoding='utf-8') as f:
                for line in f.readlines():
                    if '$dumpfile' in line:
                        found_dumpfile = True
                        break

        if not found_dumpfile:
            file_to_add = WAVES_PKG_SV_FNAME
            util.info(f'--waves arg present, no $dumpfile found, adding SV file: {file_to_add}')
            self.add_file(file_to_add)

        # register .vcd or .fst artifacts:
        util.artifacts.add_extension(
            search_paths=self.args['work-dir'], file_extension='fst',
            typ='waveform', description='Simulation Waveform FST (Fast Signal Trace) file'
        )
        util.artifacts.add_extension(
            search_paths=self.args['work-dir'], file_extension='vcd',
            typ='waveform', description='Simulation Waveform VCD (Value Change Dump) file'
        )


    def create_ext_defines_sv(self) -> None:
        '''Creates _ext_defines.sv in the work-dir, adds to front of self.files_sv'''

        if not self.defines:
            return

        if not self.args['ext-defines-sv-fname']:
            return

        fpath = os.path.join(self.args['work-dir'], self.args['ext-defines-sv-fname'])
        fpath_abs = os.path.abspath(fpath)
        if fpath_abs in self.files:
            util.warning(f'{fpath} already exists, will overwrite. Consider changing the value',
                         f'of arg: --ext-defines-sv-fname={self.args["ext-defines-sv-fname"]}')
        if not fpath.endswith('.sv') and not fpath.endswith('.v'):
            util.error(f'--ext-defines-sv-fname={self.args["ext-defines-sv-fname"]} must end with',
                       '.v or .sv')
            return

        ifdef_guard = os.path.split(self.args['ext-defines-sv-fname'])[1].upper()
        ifdef_guard = ifdef_guard.replace('.', '__').replace('/', '__')
        with open(fpath, 'w', encoding='utf-8') as f:
            f.write(f'`ifndef {ifdef_guard}\n')
            f.write(f'`define {ifdef_guard}\n')
            f.write('\n')
            for key, value in self.defines.items():
                if value is None:
                    f.write(f'`define {key}\n')
                else:
                    # if the value v is a double-quoted string, such as v='"hi"' that needs
                    # to be preserved.
                    f.write(f'`define {key} {value}\n')

            f.write('\n')
            f.write(f'`endif // {ifdef_guard}\n')

        self.files_sv.insert(0, fpath_abs)
        self.files[fpath_abs] = True
        self.files_caller_info[fpath_abs] = (
            'opencos.commands.sim::CommandSim::create_ext_defines_sv'
        )
        util.info(f'command "sim": {fpath} created for defines: {list(self.defines.keys())}')


    def update_library_map( # pylint: disable=dangerous-default-value
            self, base_search_paths: list = []
    ) -> None:
        '''
        Returns None, uses self.args['library-map'] to update:

        - self.library_map (dict)
        - self.args['sim-library']
        '''
        if not self.args['library-map']:
            return

        if not self.library_map_supported:
            tool = self.config.get('tool', None)
            util.warning(f'Command: sim, --tool={tool}, args for --library-map are not supported:',
                         f'{self.args["library-map"]}')
            return

        for libmap in self.args['library-map']:
            parts = libmap.split(':')
            lib_name = parts[0]
            lib_path = Path(':'.join(parts[1:]))
            lib_abs_path = None

            if not lib_name or not lib_path:
                util.warning(f'library-map={libmap} could not parse {lib_name=} or {lib_path=}')
                continue

            if lib_name in self.library_map:
                util.warning(
                    f'We have already mapped "{lib_name}" to {self.library_map[lib_name]}'
                )

            # We have to go looking for lib_path though
            search_files = [lib_path]
            if base_search_paths:
                for x in base_search_paths:
                    search_files.append(Path(x) / lib_path)
            for x in search_files:
                if x.exists():
                    lib_abs_path = x.resolve()
                    util.info(f'library-map={libmap} found "{lib_name}" in: {lib_abs_path}')
                    break
                util.debug(f'library-map={libmap} searching for "{lib_name}" in: {x}')


            if lib_abs_path:
                self.library_map[lib_name] = lib_abs_path
                if lib_name not in self.args['sim-library']:
                    self.args['sim-library'].append(lib_name)
            else:
                util.warning(f'library-map={libmap}, for "{lib_name}" could not find {lib_path=}')



'''opencos.commands.upload - Base class command handler for: eda upload ...

Intended to be overriden by Tool based classes (such as CommandUploadVivado, etc)
'''

import os
import re
from datetime import datetime
from pathlib import Path

from opencos.eda_base import Command, Tool
from opencos.util import Colors, debug, info, warning, safe_emoji, import_class_from_string


class CommandUpload(Command):
    '''Base class command handler for: eda upload ...

    If a --tool arg is not specified, this is the default handler for 'eda upload'
    and will attempt to choose a derived class based on the bit files found

    Tools must register their bitfile extensions in eda's CONFIG-YML, such as:

    tools:
      vivado:
        handlers:
          upload: opencos.tools.vivado.CommandUploadVivado
        upload-file-ext:
          - .bit
    '''

    command_name = 'upload'

    def __init__(self, config: dict):
        Command.__init__(self, config=config)
        self.unparsed_args = []

        self.args.update({
            'bitfile': "",
            'list-bitfiles': False,
        })

        self.supported_file_ext = self.get_supported_file_ext()
        self.ext_to_tool_map = {}

        self.str_ext = '/'.join(self.supported_file_ext).replace('.', '').upper()

        help_upload_tools = '|'.join(self.config.get('auto_tools_found', []))
        if not help_upload_tools:
            help_upload_tools = 'TOOL'

        self.args_help.update({
            'bitfile': (
                f'Tool specific {self.str_ext} files to upload (auto-detected if not specified)'
                ' If you would like see full help for a given tool, use:'
                f' {Colors.yellow}eda upload --help'
                f' {Colors.byellow}--tool={help_upload_tools}{Colors.green}'
            ),
            'list-bitfiles': (
                f'List available {self.str_ext} files.'
                ' If you would like see full help for a given tool, use:'
                f' {Colors.yellow}eda upload --help'
                f' {Colors.byellow}--tool={help_upload_tools}{Colors.green}'
            )
        })

        self.bitfiles = []

        #if not getattr(self, '_TOOL', ''):
        #    for tool, bit_exts in self.SUPPORTED_TOOLS.items():
        #        for ext in bit_exts:
        #            self.BIT_EXT_TO_TOOL[ext] = tool


    def get_supported_file_ext(self) -> list:
        '''Given the tools in config, return a list of all supported bit file extensions'''
        ret = []

        # tools to check, if this is a Tool class, then only check itself, otherwise
        # check everything from config.auto_tools_order:
        tools = []
        if isinstance(self, Tool):
            if _tool := getattr(self, '_TOOL', ''):
                tools.append(_tool) # one tool
        else:
            tools = self.config.get('auto_tools_order', {}).get('upload', [])

        for _tool in tools:
            if not _tool:
                continue

            _entry = self.config.get('tools', {}).get(_tool, {})
            if not _entry:
                continue

            if _tool not in self.config.get('tools_loaded', []):
                continue

            upload_file_ext_list = _entry.get('upload-file-ext', [])
            if isinstance(upload_file_ext_list, str):
                upload_file_ext_list = upload_file_ext_list.split()
            for ext in upload_file_ext_list:
                if ext not in ret:
                    ret.append(ext)

        return ret


    def get_tool_from_ext(self, file_ext: str) -> str:
        '''Given a valid file_ext, return the tool that should be used for it'''

        # If already known, then return it
        tool = self.ext_to_tool_map.get(file_ext, '')
        if tool:
            return tool

        tool_guesses = []
        for _tool in self.config.get('auto_tools_order', {}).get('upload', []):
            if not _tool:
                continue

            _entry = self.config.get('tools', {}).get(_tool, {})
            if not _entry:
                continue

            if _tool not in tool_guesses and \
               _tool in self.config.get('tools_loaded', []) and \
               file_ext in _entry.get('upload-file-ext', []):
                tool_guesses.append(_tool) # append, need ordered list here.

        if not tool_guesses:
            return ''
        tool = tool_guesses[0]
        self.ext_to_tool_map[file_ext] = tool # remember this choice.
        return tool #  return first one we found (from auto_tools_order)


    def process_tokens(
            self, tokens: list, process_all: bool = True, pwd: str = os.getcwd()
    ) -> list:

        self.unparsed_args = Command.process_tokens(
            self, tokens=tokens, process_all=False, pwd=pwd
        )

        if self.stop_process_tokens_before_do_it():
            return []

        self.bitfiles = self.get_list_bitfiles(display=True)

        # If someone called --list-bitfiles, stop now.
        if self.args['list-bitfiles']:
            if not self.bitfiles:
                self.error('No bitfiles found')
            return []

        sco = self._get_child_handling_class()

        if sco is None or not isinstance(sco, Tool):
            self.error('Could not find a suitable tool to process bitfiles')
            return []

        sco.unparsed_args = Command.process_tokens(
            sco, tokens=tokens, process_all=False, pwd=pwd
        )
        sco.bitfiles = self.bitfiles
        sco.create_work_dir()
        sco.do_it()
        return []

    def get_targets_or_files_from_unparsed_args(self) -> (list, list):
        '''Returns (list of targets, list of files) from unparsed args or --bitfile'''

        targets = []
        files = []
        for f in self.unparsed_args + [self.args['bitfile']]:
            if not f:
                continue
            if os.path.isfile(f):
                files.append(f)
            elif not f.startswith('-'):
                # avoid a arg
                targets.append(f)
        return targets, files


    def get_list_bitfiles(self, display: bool = True) -> list:
        '''Returns a list of bit files (ending with self.supported_file_ext)'''

        bitfiles = [] # list[Path]

        targets, files = self.get_targets_or_files_from_unparsed_args()
        targets.extend(files)

        debug(f"Looking for bitfiles in {os.path.abspath('.')=}")
        for root, _, files in os.walk("."):
            for f in files:
                if any(f.endswith(x) for x in self.supported_file_ext):
                    fullpath = os.path.abspath(Path(root) / f)
                    if os.path.isfile(fullpath) and fullpath not in bitfiles:
                        bitfiles.append(fullpath)

        matched = [] # list[Path]
        for cand in bitfiles:
            debug(f"Looking for {cand=} in {targets=}")
            passing = all(re.search(t, str(cand)) for t in targets)
            if passing:
                matched.append(cand)
                mod_time_string = datetime.fromtimestamp(
                    os.path.getmtime(cand)).strftime('%Y-%m-%d %H:%M:%S')
                tool_guess = getattr(self, '_TOOL', '')
                if not tool_guess:
                    ext = os.path.splitext(cand)[1]
                    tool_guess = self.get_tool_from_ext(file_ext=ext)
                if tool_guess:
                    tool_guess = f'({tool_guess})'
                if display:
                    info(
                        f"{safe_emoji('⏩ ')}Found matching bitfile {tool_guess}:",
                        f"{Colors.cyan}{mod_time_string}{Colors.off} :",
                        f"{Colors.byellow}{cand}"
                    )

        if display and not matched:
            if self.args['list-bitfiles']:
                warning(f'{safe_emoji("❕ ")}--list-bitfiles: no {self.str_ext} found that matched',
                        f'{targets}')
            else:
                warning(f'{safe_emoji("❕ ")} Searched for bitfiles with {self.str_ext}: none found',
                        f'that matched {targets}')

        return matched


    def _get_child_handling_class(self) -> object:
        '''Returns a class handle of a child to process this, which should be a Tool class

        if no appropriate child is found, returns self.
        '''

        if isinstance(self, Tool):
            # We're already a tool handling class.
            return self

        tools_found = set()
        for bitfile in self.bitfiles:
            ext = os.path.splitext(bitfile)[1]
            tool_guess = self.get_tool_from_ext(file_ext=ext)
            if tool_guess:
                tools_found.add(tool_guess)
            else:
                warning(f'For bitfile {bitfile} no tool found for it')
                return self

        if not tools_found:
            # Probably not an error?
            warning(f'No tools found to process bitfiles: {self.bitfiles}')
            return self

        if len(tools_found) > 1:
            warning(f'More than one tool found ({tools_found}) to to process bitfiles:',
                         f'{self.bitfiles}')
            return self


        tool = tools_found.pop() # only item in set
        # Do we have a handler for this in our config?
        # TODO(drew): if present in config.auto_tools_found for 'upload', we'd prefer to
        # use that order. There's already an above warning about this.
        if tool in self.config.get('tools_loaded', []):
            cls_str, cls, sco = None, None, None
            tool_cfg = self.config.get('tools', {}).get(tool, {})

            if tool_cfg:
                cls_str = tool_cfg.get('handlers', {}).get(self.command_name, None)
            if cls_str:
                cls = import_class_from_string(cls_str)
            if cls and issubclass(cls, Command):
                info(f'For found bitfiles, can use tool={tool} and handler {cls}')
                sco = cls(config=self.config)
                return sco

        warning(f'No handler found for tool={tool} to process bitfiles: {self.bitfiles}')
        debug(f'config -- tools_loaded: {self.config["tools_loaded"]}')
        debug(f'config -- tools for tool: {self.config["tools"].get(tool, "")}')
        return self



''' opencos.tools.riviera - Used by opencos.eda for sim/elab commands w/ --tool=riviera.

Contains classes for ToolRiviera, CommandSimRiviera, CommandElabRiviera.
'''

# pylint: disable=too-many-ancestors
# pylint: disable=R0801 # (duplicate code in derived classes, such as if-condition return.)

import argparse
import glob
import os
from pathlib import Path
import shutil
import tempfile
import time

from opencos import util
from opencos.commands import CommandFList, CommandWaves
from opencos.commands.coverage import CommandCoverage
from opencos.files import safe_shutil_which, safe_output_file
from opencos.commands.sim import WAVES_PKG_SV_FNAME
from opencos.tools.questa_common import ToolQuesta, CommonSimQuesta
from opencos.utils import status_constants, vsim_helper
from opencos.utils.str_helpers import sanitize_defines_for_sh

from opencos.tools.riviera_coverage import RivieraSimCoverageAddOn, \
    get_acdb_output_file_from_tcl_line

class ToolRiviera(ToolQuesta):
    '''ToolRiviera used by opencos.eda for --tool=riviera'''

    _TOOL = 'riviera'
    _EXE = 'vsim'
    use_vopt = False
    uvm_versions = set()
    uvm_home_dirs = {} # key (str item in uvm_versions): value (str path where uvm_pkv.sv exists)


    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        path = self.config.get('auto_tools_found', {}).get(self._TOOL, '')
        if not path and self._EXE == 'vsim':
            # otherwise, use the vsim helper:
            path = vsim_helper.lookup_exe(self._TOOL)
        if not path:
            # otherwise, fallback to shutil which:
            path = safe_shutil_which(self._TOOL)
        if not path:
            if self.get_version_allow_certain_command_if_tool_not_installed():
                return self._VERSION
            self.error(f"{self._EXE} not in path, need to setup or add to PATH")
            util.debug(f"{path=}")
        else:
            self.sim_exe = path
            self.sim_exe_base_path, _ = os.path.split(path)
            self.waveviewer_exe = os.path.join(self.sim_exe_base_path, 'vsim')

        self._VERSION = vsim_helper.lookup_version(self._TOOL)

        # Get the UVM versions in the install directory. Note this may run
        # more than once, so only do this if self.uvm_versions not yet set:
        riviera_path, _ = os.path.split(self.sim_exe_base_path)
        vlib_path = os.path.join(riviera_path, 'vlib')
        if not self.uvm_versions and os.path.isdir(vlib_path):
            for item in os.listdir(vlib_path):
                # uvm-1.1, uvm-1.1d - so don't pick anything > 9 chars (uvm-M.mRr)
                if item.startswith('uvm-') and '1800' not in item and len(item) <= 9:
                    uvm_ver = item[4:]
                    self.uvm_versions.add(uvm_ver)
                    # While we're at it, make note of location of uvm_pkg.sv?
                    if os.path.isfile(os.path.join(vlib_path, item, 'src', 'uvm_pkg.sv')):
                        self.uvm_home_dirs[uvm_ver] = os.path.join(vlib_path, item, 'src')

        return self._VERSION


class CommandWavesRiviera(CommandWaves, ToolRiviera):
    '''Handler for: eda waves --tool=riviera'''

    def __init__(self, config: dict):
        CommandWaves.__init__(self, config=config)
        ToolRiviera.__init__(self, config=self.config)

    def do_it(self) -> None:
        # If you don't like the self.waveviewer_exe, use:
        #    os.path.join(self.sim_exe_base_path, 'vsim') or whichever you prefer
        if not self.waveviewer_exe:
            self.error(f'Unable to find the exe for --tool={self._TOOL} for command waves',
                       'check "eda --help" or PATH for vsim')
        command_list = [
            self.waveviewer_exe, self.wave_file
        ]
        self.exec(os.path.dirname(self.wave_file), command_list)


class CommandSimRiviera(CommonSimQuesta, ToolRiviera):
    '''CommandSimRiviera is a command handler for: eda sim --tool=riviera

    From CommonSimQuesta (and opencos.commands.sim.CommandSim), the common entrypoint is:
        -- process_tokens()
             --> run commands from DEPS
             --> run_argparser_on_list() -- populate self.args
             --> do_it() / run()
                 --> prepare_compile()
                       --> get_*_command_list()
                 --> compile(), elaborate(), simulate()
             --> run_post_tool_dep_commands()
             --> report pass/fail
             --> return unparsed args back to eda.py.


    When using --coverage:
    self.coverage_obj (opencos.tools.riviera_coverage.RiveriaSimCoverageAddOn) is used to
    provide addtional coverage features. See riviera_coverage.py for full details.
        - add_coverage_args_to_handler() -- applied with --coverage arg present
        - calculate_coverage_delta(metrics, metrics)
        - check_delta_thresholds(delta metrics)
        - display_coverage_delta(delta metrics)
        - display_coverage_summary(metrics dict)
        - sim_post_simulate_coverage() -- called with --coverage to create better summaries
              outside of the Riviera vsim.do TCL flow
        - generate_baseline_report(acdb fpath) - gets temporary directory with ACDB reports
        - generate_coverage_json(cov.txt, json output, eda target)
        - generate_delta_json(delta metrics, path, json output)
        - get_toggle_coverage_commands() -- generates vsim TCL commands, used prior to
              "run -a" to apply toggle coverage items (after vlog, vsim steps, prior to run step)
        - get_covergroup_coverage_commands() -- generates vsim TCL command to apply covergroup
              (bins) after vlog and vsim steps, prior to run step.
        - parse_coverage_metrics(cov.txt) -- returns metrics dict
    '''

    library_map_supported: bool  = True
    use_add_on_coverage_ast: bool = False

    def __init__(self, config: dict):
        CommonSimQuesta.__init__(self, config=config)

        # This technically inherits ToolQuesta and ToolRiviera, but ToolRiviera will
        # override get_versions(). Just be careful on using issinstance(cls_obj, ToolQuesta),
        # and instead use self._TOOL, or self.args['tool']
        ToolRiviera.__init__(self, config=self.config)

        self.shell_command = os.path.join(self.sim_exe_base_path, 'vsim')
        self.starter_edition = False


        self.args.update({
            'tool': self._TOOL, # override
            'gui': False,
            'waves-fst': True,
            'waves-vcd': False,
            'custom-run-tcl': '',
            'coverage-tcl': '',
            'coverage-covergroups': False,
            'ext-defines-sv-fname': '_ext_defines.sv',
            'license-queue': True,
            'license-queue-value': '30',
        })
        self.args_help.update({
            'waves-fst': (
                '(Default True) If using --waves, apply simulation runtime arg +trace.'
                ' Note that if you do not have SV code using $dumpfile, eda will add'
                ' _waves_pkg.sv to handle this for you with +trace runtime plusarg.'
            ),
            'waves-vcd': 'If using --waves, apply simulation runtime arg +trace=vcd',
            'waves': 'Save a .asdb offline wavefile, can be used with --waves-fst or --waves-vcd',
            'coverage': (
                'See --coverage with --help for many additional --coverage-* args'
                ' many of which require external tool "slang". --coverage on its own will'
                ' run basic Riviera coverage, save a work.acdb and cov.txt report, such as:'
                ' --coverage-summary'
            ),
            'coverage-tcl': (
                'bring your own .tcl file to run in Riviera (vsim) for coverage. The default'
                ' tcl steps are (from tool config in --config-yml):   '
            ) + '; '.join(self.tool_config.get('simulate-coverage-tcl', [])),
            'uvm': (
                'Attempts to support UVM. For Riviera, this adds "-uvmver NUMBER -dbg" to vlog.f.'
                ' You can choose your uvmver using eda arg --uvm-version=NUMBER.'
                ' Also adds +access +r to vopt/vsim. There is no -l or -L library modifications,'
                ' we rely on Riviera to handle this internally based on running: vlog -uvm'
            ),
            'license-queue': (
                'Sets, via TCL, in-tool variables for LICENSE_QUEUE=30, which should set Riviera'
                ' to wait for licenses instead of immediately exiting with error. If unset'
                ' we also check ENV var LICENSE_QUEUE=VALUE, and if present that is used instead'
            ),
            'license-queue-value': (
                'Value used (LICENSE_QUEUE=VALUE) if --license-queue is set, which takes higher'
                ' precedence than ENV var LICENSE_QUEUE=VALUE'
            ),
            'custom-run-tcl': (
                'Bring your own .tcl file that will be included after the "vsim" step, as a'
                ' replacement for the default "run -all"'
            )
        })

        if self.uvm_versions:
            # set default to latest version:
            self.args['uvm-version'] = sorted(self.uvm_versions)[-1]

        self.args_kwargs.update({
            'uvm-version': { 'choices': list(self.uvm_versions) },
        })

        # Note this will NOT update self.args yet:
        self.coverage_obj = RivieraSimCoverageAddOn(handler_obj=self)


    def run_argparser_on_list(
            self, tokens: list, parser_arg_list=None, apply_parsed_args: bool = True
    ) -> (argparse.Namespace, list):
        '''
        Override on Command.run_argparser_on_list - to conditionally add
        args from self.coverage_obj if --coverage or --coverage-* is set
        '''

        # This is a little inefficient, we could instead make a customer argparser just
        # for --coverage, but this also works.
        # Run argparser w/out appyling to fish out --coverage:
        if not self.coverage_obj.have_added_coverage_args_to_handler:
            parsed, unparsed = super().run_argparser_on_list(
                tokens=tokens, parser_arg_list=parser_arg_list, apply_parsed_args=False
            )

            add_on_args = parsed.coverage
            if not add_on_args:
                # Be nice and do this if user ran with --coverage-summary but forgot to
                # add --coverage:
                add_on_args = any(x.startswith('--coverage-') for x in unparsed)

            if add_on_args:
                # Add args from riviera_coverage b/c --coverage, note that self.coverage_obj
                # should only do this once.
                self.coverage_obj.add_coverage_args_to_handler()


        # Run the argparser on the original tokens, applying them this time (if caller
        # wanted them applied)
        parsed, unparsed = super().run_argparser_on_list(
            tokens=tokens, parser_arg_list=parser_arg_list, apply_parsed_args=apply_parsed_args
        )

        self.use_add_on_coverage_ast = self.coverage_obj.any_add_on_coverage_ast_args_set()
        util.debug(f'CoverageAddOn - {self.use_add_on_coverage_ast=}')
        return parsed, unparsed


    # Note: many of these we follow the same flow as CommandSimModelsimAse:
    # do_it, prepare_compile, compile, elaborate, simulate

    def compile(self):
        '''Override for CommandSimModelsimAse.compile() so we can set our own must_strings'''

        if self.args['stop-before-compile']:
            # don't run anything, save everyting we've already run in _prep_compile()
            return
        if self.args['stop-after-compile']:
            vsim_command_lists = self.get_compile_command_lists()
            self.run_commands_check_logs(vsim_command_lists, log_filename='sim.log',
                                         must_strings=['Compile success 0 Errors'],
                                         use_must_strings=False,
                                         sim_step='compile')

    def simulate(self):
        '''Override to add coverage reporting after simulation completes'''

        # Run normal sim.CommandSim.simulate() step:
        super().simulate()

        if self.args.get('coverage', False):
            # Run RiveraCoverageAddOn.sim_post_simulate_coverage() step:
            self.coverage_obj.sim_post_simulate_coverage()
            # Add artifacts
            util.artifacts.add_extension(
                search_paths=self.args['work-dir'], file_extension='acdb',
                typ='coverage_database',
                description='Riviera Coverage database binary file'
            )

    def get_compile_command_lists(self, **kwargs) -> list:
        # This will also set up a compile.
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-l', 'sim.log', '-do', 'vsim_vlogonly.do'
        ]
        return [vsim_command_list]

    def get_elaborate_command_lists(self, **kwargs) -> list:
        # This will also set up a compile, for vlog + vsim (0 time)
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-l', 'sim.log', '-do', 'vsim_lintonly.do',
        ]
        return [vsim_command_list]

    def get_simulate_command_lists(self, **kwargs) -> list:
        # This will also set up a compile, for vlog + vsim (with run -all)
        vsim_command_list = [
            self.sim_exe,
            '' if self.args['gui'] else '-c',
            '-l', 'sim.log', '-do', 'vsim.do',
        ]
        return [vsim_command_list]

    def get_post_simulate_command_lists(self, **kwargs) -> list:
        return []


    def write_vlog_dot_f( # pylint: disable=too-many-branches
            self, filename: str = 'vlog.f'
    ) -> None:
        '''Returns none, creates filename (str) for a vlog.f'''
        vlog_dot_f_lines = []

        # Add compile args from config.tool.riviera
        vlog_dot_f_lines += self.tool_config.get(
            'compile-args',
            '-sv -input_ports net').split()

        # Add waivers from config.tool.riviera, convert to warning:
        for waiver in self.tool_config.get('compile-waivers', []) + \
                self.args['compile-waivers']:
            vlog_dot_f_lines += [f'-err {waiver} W1']

        vlog_dot_f_fname = filename
        vlog_dot_f_fpath = os.path.join(self.args['work-dir'], vlog_dot_f_fname)

        if self.args['uvm']:
            vlog_dot_f_lines.extend([
                f'-uvmver {self.args["uvm-version"]}',
                '-dbg'
            ])

        # Note, in Riviera we have to put sim libs in vsim call (with -L) and in vlog (with -l)
        vlog_dot_f_lines.extend([f'-l {x}' for x in self.args['sim-library']])

        for value in self.incdirs:
            vlog_dot_f_lines += [ f"+incdir+{value}" ]

        # For external defines, we will create a file in the work-dir named
        # _ext_defines.sv, it will be the first file in the file list, and it
        # will be valid Verilog/SV, not a list of +define+Name[=Value]
        if self.args['ext-defines-sv-fname']:
            self.create_ext_defines_sv()
        else:
            for k,v in self.defines.items():
                if v is None:
                    vlog_dot_f_lines += [ f'+define+{k}' ]
                else:
                    # if the value v is a double-quoted string, such as v='"hi"', the
                    # entire +define+NAME="hi" needs to wrapped in double quotes with the
                    # value v double-quotes escaped: "+define+NAME=\"hi\""
                    if isinstance(v, str) and v.startswith('"') and v.endswith('"'):
                        str_v = v.replace('"', '\\"')
                        vlog_dot_f_lines += [ f'"+define+{k}={str_v}"' ]
                    else:
                        # Generally we should only support int and str python types passed as
                        # +define+{k}={v}, but also for SystemVerilog plusargs
                        vlog_dot_f_lines += [ f'+define+{k}={sanitize_defines_for_sh(v)}' ]



        vlog_dot_f_lines += self.args['compile-args']
        if self.args['coverage']:
            vlog_dot_f_lines += self.tool_config.get('compile-coverage-args', '').split()

        vlog_dot_f_lines += [
            ] + list(self.files_sv) + list(self.files_v)

        if not self.files_sv and not self.files_v:
            if not self.args['stop-before-compile']:
                self.error(f'{self.target=} {self.files_sv=} and {self.files_v=} are empty,',
                           'cannot create a valid vlog.f')

        with open(vlog_dot_f_fpath, 'w', encoding='utf-8') as f:
            f.writelines(line + "\n" for line in vlog_dot_f_lines)



    def write_vsim_dot_do( # pylint: disable=too-many-branches,too-many-locals,too-many-statements
            self, dot_do_to_write: list
    ) -> None:
        '''Writes files(s) based on dot_do_to_write(list of str)

        list arg values can be empty (all) or have items 'all', 'sim', 'lint', 'vlog'.'''

        vsim_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim.do')
        vsim_lintonly_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim_lintonly.do')
        vsim_vlogonly_dot_do_fpath = os.path.join(self.args['work-dir'], 'vsim_vlogonly.do')

        sim_plusargs_str = self._get_sim_plusargs_str()
        vsim_ext_args = ' '.join(self.args.get('sim-args', []))

        if self.args['waves'] and '+trace' not in sim_plusargs_str:
            if self.args.get('waves-vcd', False):
                sim_plusargs_str += ' +trace=vcd'
            elif self.args.get('waves-fst', False):
                sim_plusargs_str += ' +trace'

        voptargs_str = self.tool_config.get('elab-args', '')
        voptargs_str += ' '.join(self.args.get('elab-args', []))
        if self.args['gui'] or self.args['waves'] or self.args['coverage']: # \
           #or self.args['waves-asdb']:
            voptargs_str += self.tool_config.get('simulate-waves-args',
                                                 '+accb +accr +access +r+w')
        if self.args['coverage']:
            voptargs_str += ' ' + self.tool_config.get('coverage-args', '')

        if self.args['uvm']:
            if all(x not in voptargs_str for x in ('+access +r', '+access +w+r')):
                voptargs_str += ' +access +r'

        # parameters, for Riviera we need to use the leading forward slash option
        # aka -G/Name=Value, if our parameter name doesn't already have path information.
        # If you do -GName=Value, all parameters in the hierachy with Name will be set,
        # so to only do top level parameters, if hierarchy isn't mentioned in the Name,
        # would need to do -G/{self.args['top']}/Name=Value.
        if self.parameters:
            voptargs_str += ' ' + ' '.join(
                self.process_parameters_get_list(
                    arg_prefix='-G', hier_delimiter='/', top_hier_str=f'/{self.args["top"]}/'
                )
            )

        # Note, in Riviera we have to put sim libs in vsim call (with -L) and in vlog (with -l)
        # we previously updated self.args['sim-library'] with self.update_library_map()
        vmap_lines_str = self.get_vmap_lines_str()
        vsim_libs = ' '.join([f'-L {x}' for x in self.args['sim-library']])

        vsim_one_liner = (
            "vsim"
            f" -sv_seed {self.args['seed']} {sim_plusargs_str}"
            f" {voptargs_str} {vsim_ext_args} {vsim_libs} work.{self.args['top']}"
        )

        if self.args['waves'] and WAVES_PKG_SV_FNAME in self.files:
            # kind of a hack, but add _waves_module (from _waves_pkg.sv) if that file
            # was in our files list.
            vsim_one_liner += ' work._waves_module'

        for l in self.args['add-top-library']:
            vsim_one_liner += f' {l}'

        vsim_one_liner = vsim_one_liner.replace('\n', ' ') # needs to be a one-liner

        dot_do_start_lines = []

        if self.args['license-queue']:
            dot_do_start_lines.append(
                f'set LICENSE_QUEUE {self.args.get("license-queue-value", 30)};'
            )
        elif value := os.environ.get('LICENSE_QUEUE', ''):
            dot_do_start_lines.append(f'set LICENSE_QUEUE {value};')

        dot_do_start_lines.extend([
            "if {[file exists work]} { vdel -all work; }",
            "vlib work;",
            "set qc 30;",
            vmap_lines_str,
        ])


        vsim_vlogonly_dot_do_lines = dot_do_start_lines + [
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -code 20 -force;",
            "    }",
            "}",
            "if {[batch_mode]} {",
            "    quit -code 0 -force;",
            "}",
        ]

        vsim_lintonly_dot_do_lines = dot_do_start_lines + [
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -code 20 -force;",
            "    }",
            "}",
            "if {[catch { " + vsim_one_liner + " } result] } {",
            "    echo \"Caught $result\";",
            "    if {[batch_mode]} {",
            "        quit -code 19 -force;",
            "    }",
            "}",
            "if {[batch_mode]} {",
            "    quit -code 0 -force;",
            "}",
        ]

        vsim_dot_do_lines = dot_do_start_lines + [
            "if {[catch {vlog -f vlog.f} result]} {",
            "    echo \"Caught $result \";",
            "    if {[batch_mode]} {",
            "        quit -code 20 -force;",
            "    }",
            "}",
            "if {[catch { " + vsim_one_liner + " } result] } {",
            "    echo \"Caught $result\";",
            "    if {[batch_mode]} {",
            "        quit -code 19 -force;",
            "    }",
            "}",
        ]

        #
        # --coverage features:
        #
        # Insert toggle or covergroup coverage commands AFTER vsim but BEFORE run -all
        # This is required by Riviera which cannot enable toggle coverage globally
        #
        if self.use_add_on_coverage_ast:

            # get an AST and read it (will be in self.coverage_obj.slang_ast.ast_data)
            # This also handled UVM for slang:
            # The saved .sh script is in work-dir: self.coverage_obj.slang_ast.slang_sh_fname
            self.coverage_obj.run_slang_for_ast_auto()
            # Set this for CommandSim.sim to be aware of in write_sh_scripts_to_work_dir(..)
            self.has_pre_compile_slang_sh = self.coverage_obj.slang_ast.slang_sh_fname


            if self.args.get('coverage-toggle', False):
                vsim_dot_do_lines.extend(self.coverage_obj.get_toggle_coverage_commands())


        if fpath := self.args.get('custom-run-tcl', ''):
            if not os.path.exists(self.args['custom-run-tcl']):
                self.error(f'--custom-run-tcl={fpath} file does not exist')
            with open(fpath, 'r', encoding='utf-8', errors='replace') as f:
                vsim_dot_do_lines.extend(['', f"## Included from: --custom-run-tcl={fpath}"])
                for line in f.readlines():
                    vsim_dot_do_lines.append(line.rstrip('\n'))
                vsim_dot_do_lines.extend([f"## End include from: (--custom-run-tcl={fpath}", ''])
        else:
            vsim_dot_do_lines.append("run -all;")

        if self.args['coverage']:
            will_run_sim = not(any(self.args.get(x, None) for x in ('stop-before-compile',
                                                                    'stop-after-elaborate',
                                                                    'stop-after-compile')) or
                               self.command_name in ('elab', 'lint'))
            if self.args['coverage-tcl']:
                tclfile = os.path.abspath(self.args['coverage-tcl'])
                if not os.path.isfile(tclfile):
                    self.error(f'--coverage-tcl file not found: {tclfile}',
                               error_code=status_constants.EDA_COMMAND_OR_ARGS_ERROR)
                vsim_dot_do_lines += [
                    f'source {tclfile}'
                ]
                # Unfortunately we won't know what's in the user supplied coverage-tcl file, so
                # for artifacts we have to open it and try to fish them out.
                # Also only add these *if* we're running the simulation
                if will_run_sim:
                    with open(tclfile, 'r', encoding='utf-8') as f:
                        for line in f.readlines():
                            get_acdb_output_file_from_tcl_line(
                                line, work_dir=self.args['work-dir'], add_to_artifacts=True
                            )

            else:
                tcl_lines = self.tool_config.get('simulate-coverage-tcl', [])

                if will_run_sim:
                    # default TCL for coverage, similarly go through the lines for ACDB artifacts
                    # Also only add these *if* we're running the simulation
                    for line in tcl_lines:
                        get_acdb_output_file_from_tcl_line(
                            line, work_dir=self.args['work-dir'], add_to_artifacts=True
                        )

                vsim_dot_do_lines += tcl_lines


        vsim_dot_do_lines += [
            "if {[batch_mode]} {",
            "    quit -code 0 -force;",
            "}",
        ]

        write_all = len(dot_do_to_write) == 0 or 'all' in dot_do_to_write
        if write_all or 'sim' in dot_do_to_write:
            with open(vsim_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_dot_do_lines)

        if write_all or 'lint' in dot_do_to_write:
            with open(vsim_lintonly_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_lintonly_dot_do_lines)

        if write_all or 'vlog' in dot_do_to_write:
            with open(vsim_vlogonly_dot_do_fpath, 'w', encoding='utf-8') as f:
                f.writelines(line + "\n" for line in vsim_vlogonly_dot_do_lines)


    def _get_vsim_suppress_list_str(self) -> str:
        vsim_suppress_list = []
        # Add waivers from config.tool.modelsim_ase:
        for waiver in self.tool_config.get(
                'simulate-waivers', [
                    #defaults: none
                ]) + self.args['sim-waivers']:
            vsim_suppress_list += ['-filter', str(waiver)]

        return ' '.join(vsim_suppress_list)


    def update_library_map( # pylint: disable=dangerous-default-value
            self, base_search_paths: list = []
    ) -> None:
        '''Override from sim::CommandSim

        We add some common places to look relative to the Riviera install paths
        '''
        bsp = [] + base_search_paths
        bsp += [
            Path(self.sim_exe_base_path),
            Path(self.sim_exe_base_path) / '..',
            Path(self.sim_exe_base_path) / '..' / 'vlib',
        ]
        super().update_library_map(base_search_paths=bsp)

    def get_vmap_lines_str(self) -> str:
        '''Returns an empty str, or single str in the form:

        vmap -global LibraryName  AbsolutePathToADotLibFile
        vmap -global LibraryName2 AbsolutePathToAnotherDotLibFile
        ...
        uses self.args['library-map'], updates self.args['sim-library']
        '''
        if not self.library_map:
            return ''

        lines = []
        for lib_name, lib_file_path in self.library_map.items():
            lines.append(f'vmap -global {lib_name} {lib_file_path}')

        return '\n'.join(lines)


class CommandElabRiviera(CommandSimRiviera):
    '''CommandElabRiviera is a command handler for: eda elab --tool=riviera'''

    command_name = 'elab'

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-elaborate'] = True


class CommandLintRiviera(CommandSimRiviera):
    '''CommandLintRiviera is a command handler for: eda lint --tool=riviera'''

    command_name = 'lint'

    def __init__(self, config:dict):
        super().__init__(config)
        self.args['stop-after-compile'] = True
        self.args['stop-after-elaborate'] = True


class CommandFListRiviera(CommandFList, ToolRiviera):
    '''CommonFListQuesta is a command handler for: eda flist --tool=riviera'''

    def __init__(self, config: dict):
        CommandFList.__init__(self, config=config)
        ToolRiviera.__init__(self, config=self.config)


class CommandCoverageRiviera(CommandCoverage, ToolRiviera):
    '''Riviera-specific coverage merge and compare handler'''

    def __init__(self, config: dict):
        CommandCoverage.__init__(self, config=config)
        ToolRiviera.__init__(self, config=self.config)

        # Note this will NOT update self.args yet:
        self.coverage_obj = RivieraSimCoverageAddOn(handler_obj=self)


    def do_it(self) -> None:
        '''Execute the coverage operation based on subcommand'''

        if self.subcommand == 'merge':
            self.run_merge()
        elif self.subcommand == 'compare':
            self.run_compare()
        else:
            self.error(f"Subcommand '{self.subcommand}' not implemented for Riviera",
                      error_code=status_constants.EDA_COMMAND_OR_ARGS_ERROR)

    def run_merge( # pylint: disable=too-many-locals,too-many-branches,too-many-statements
            self
    ) -> None:
        '''Execute the coverage merge operation'''

        # If no path on the output arg, put it in work-dir:
        self.create_work_dir()
        output_file = safe_output_file(
            work_dir=self.args['work-dir'], default_fname='merged.acdb',
            user_fpath=self.args['output']
        )
        output_dir = os.path.dirname(output_file)

        tcl_script_path = os.path.join(self.args['work-dir'], 'merge_coverage.tcl')
        merge_command = (
            'acdb merge ' + ' '.join(f'-i {f}' for f in self.coverage_files) + f' -o {output_file}'
        )
        tcl_lines = [
            '# Auto-generated TCL script for coverage merging',
            merge_command
        ]

        if self.args['coverage-report']:
            report_txt = os.path.join(output_dir, self.args['coverage-report-txt'])
            tcl_lines.extend([
                f'acdb report -i {output_file} -txt -o {report_txt}',
                f'acdb report -i {output_file} -txt -o' + \
                f' {report_txt.replace(".txt", "_covergroups.txt")} -show covergroups',
            ])
            if self.args['coverage-report-html']:
                tcl_lines.append(
                    f'acdb report -i {output_file} -html -o '
                    f'{os.path.join(output_dir, self.args["coverage-report-html"])}'
                )

        tcl_lines.append('exit')

        with open(tcl_script_path, 'w', encoding='utf-8') as f:
            f.write('\n'.join(tcl_lines) + '\n')

        util.info(f"Generated TCL script: {tcl_script_path}")

        util.info(f"Merging {len(self.coverage_files)} coverage databases...")
        self.exec(self.args['work-dir'], [self.sim_exe, '-c', '-do', 'merge_coverage.tcl'])

        if not os.path.exists(output_file):
            self.error(f"Coverage merge failed: output file not created: {output_file}",
                       error_code=status_constants.EDA_COVERAGE_TOOL_COMMAND_FAIL)

        util.info(f"Coverage merge successful: {output_file}")
        if not self.args['coverage-report']:
            return

        report_txt = os.path.join(output_dir, self.args['coverage-report-txt'])
        if not os.path.exists(report_txt):
            return

        util.info(f"Coverage report: {report_txt}")

        metrics = self.coverage_obj.parse_coverage_metrics(report_txt)
        if metrics and self.args.get('coverage-summary', True):
            self.coverage_obj.display_coverage_summary(metrics)

        if self.args.get('coverage-json', ''):
            json_path = str(Path(self.args['coverage-json']).absolute())
            self.coverage_obj.generate_coverage_json(
                report_txt, json_path, target_name='merged_coverage'
            )

        baseline_path = ''
        if self.args.get('coverage-baseline', ''):
            baseline_path = os.path.abspath(self.args['coverage-baseline'])
        elif self.args.get('coverage-baseline-auto', False):
            baseline_dir_abs = str(
                Path(self.args.get('coverage-baseline-dir', '.coverage_baselines')).absolute()
            )
            latest_link = os.path.join(baseline_dir_abs, 'merged_latest.acdb')
            if os.path.exists(latest_link):
                baseline_path = os.path.realpath(latest_link)
            else:
                candidates = sorted(
                    glob.glob(os.path.join(baseline_dir_abs, 'merged_*.acdb')),
                    key=os.path.getmtime, reverse=True
                )
                baseline_path = candidates[0] if candidates else ''

        if baseline_path and os.path.exists(baseline_path):
            util.info(f"Comparing merged coverage against baseline: {baseline_path}")
            temp_dir, baseline_report_temp = self.coverage_obj.generate_baseline_report(
                baseline_path
            )
            if baseline_report_temp and os.path.exists(baseline_report_temp):
                baseline_metrics = self.coverage_obj.parse_coverage_metrics(
                    baseline_report_temp
                )
                if baseline_metrics and metrics:
                    delta_metrics = self.coverage_obj.calculate_coverage_delta(
                        baseline_metrics, metrics
                    )
                    if self.args.get('coverage-show-delta', False):
                        self.coverage_obj.display_coverage_delta(delta_metrics)
                    if delta_metrics and \
                       not self.coverage_obj.check_delta_thresholds(delta_metrics):
                        self.error(
                            "Merged coverage delta thresholds not met",
                            error_code=status_constants.EDA_COVERAGE_REQUIREMENT_NOT_MET
                        )
                    if self.args.get('coverage-delta-json', ''):
                        delta_json_path = str(Path(self.args['coverage-delta-json']).absolute())
                        self.coverage_obj.generate_delta_json(
                            delta_metrics, baseline_path, delta_json_path
                        )
            temp_dir.cleanup()
        elif baseline_path:
            util.warning(f"Baseline not found: {baseline_path}")

        if self.args.get('coverage-save-baseline', False) and os.path.exists(output_file):
            baseline_dir_abs = str(
                Path(self.args.get('coverage-baseline-dir', '.coverage_baselines')).absolute()
            )
            os.makedirs(baseline_dir_abs, exist_ok=True)
            baseline_name = f"merged_{time.strftime('%Y%m%d_%H%M%S')}.acdb"
            baseline_path_save = os.path.join(baseline_dir_abs, baseline_name)
            shutil.copy2(output_file, baseline_path_save)
            util.info(f"Saved merged baseline: {baseline_path_save}")
            latest_link = os.path.join(baseline_dir_abs, 'merged_latest.acdb')
            if os.path.exists(latest_link) or os.path.islink(latest_link):
                os.remove(latest_link)
            os.symlink(baseline_name, latest_link)
            util.info(f"Updated latest merged baseline link: {latest_link}")


    def run_compare(self) -> None:
        '''Execute coverage comparison between two ACDB files'''

        if len(self.coverage_files) != 2:
            util.warning('Need exactly two coverage files, however there are',
                         f'{len(self.coverage_files)} from positional args')
        if not self.coverage_files:
            self.error('No coverage files to compare, exiting',
                       error_code=status_constants.EDA_COMMAND_OR_ARGS_ERROR)

        self.create_work_dir()
        baseline_acdb, current_acdb = self.coverage_files[0], self.coverage_files[-1]
        util.info(f"Baseline: {baseline_acdb}")
        util.info(f"Current:  {current_acdb}")

        temp_dir = tempfile.TemporaryDirectory( # pylint: disable=consider-using-with
            dir=self.args['work-dir'],
            prefix='coverage_compare_'
        )

        baseline_report = os.path.join(temp_dir.name, 'baseline_report.txt')
        with open(os.path.join(temp_dir.name, 'baseline_report.tcl'), 'w', encoding='utf-8') as f:
            f.write(
                f'acdb report -i {baseline_acdb} -txt -o {baseline_report}\nexit\n'
            )

        self.exec(
            temp_dir.name,
            [self.sim_exe, '-c', '-do', 'baseline_report.tcl']
        )
        if not os.path.exists(baseline_report):
            self.error(
                f"Failed to generate baseline report from {baseline_acdb} at: {baseline_report}",
                error_code=status_constants.EDA_COVERAGE_TOOL_COMMAND_FAIL)

        current_report = os.path.join(temp_dir.name, 'current_report.txt')
        with open(os.path.join(temp_dir.name, 'current_report.tcl'), 'w', encoding='utf-8') as f:
            f.write(f'acdb report -i {current_acdb} -txt -o {current_report}\nexit\n')

        self.exec(
            temp_dir.name,
            [self.sim_exe, '-c', '-do', 'current_report.tcl']
        )
        if not os.path.exists(current_report):
            self.error(
                f"Failed to generate current report from {current_acdb}",
                error_code=status_constants.EDA_COVERAGE_TOOL_COMMAND_FAIL
            )

        util.info("Parsing coverage reports...")
        baseline_metrics = self.coverage_obj.parse_coverage_metrics(baseline_report)
        current_metrics = self.coverage_obj.parse_coverage_metrics(current_report)

        if not baseline_metrics:
            util.warning("No coverage metrics found in baseline report")
        if not current_metrics:
            util.warning("No coverage metrics found in current report")

        if baseline_metrics and current_metrics:
            delta_metrics = self.coverage_obj.calculate_coverage_delta(
                baseline_metrics, current_metrics
            )
            if self.args.get('coverage-show-delta', False):
                self.coverage_obj.display_coverage_delta(delta_metrics)
            if json_path := self.args.get('coverage-delta-json', ''):
                json_path = str(Path(json_path).absolute())
                self.coverage_obj.generate_delta_json(
                    delta_metrics, baseline_acdb, json_path
                )
            util.info("Coverage comparison complete")
        else:
            util.warning("Could not complete coverage comparison due to missing metrics")

        # If we made it this far, cleanup the tempdir, otherwise leave it for debug:
        temp_dir.cleanup()



#!/usr/bin/env python3

"""
CLI script to process timing report from OpenSTA and generate timing metrics CSV.

This script parses the output from report_checks and calculates:
- Total Paths
- WNS (Worst Negative Slack)
- TNS (Total Negative Slack)
- NFE (Number of Failing Endpoints)
- Sum of Arrival/Required Ratio
- Individual path slacks
"""

import re
import sys
import os
import csv


def parse_timing_report(report_file: str) -> list:
    """
    Parse timing report file and extract timing metrics.

    Expected format:
    Endpoint                                 Delay     Delay     Slack
    ------------------------------------------------------------------
    _42791_/D (sky130_fd_sc_hs__dfrtp_1)    9.8396   12.9553   -3.1157 (VIOLATED)
    """

    paths = []

    with open(report_file, 'r', encoding='utf-8', errors='replace') as f:
        lines = f.readlines()

    # Find the data section (after the header line with dashes)
    in_data_section = False

    for line in lines:
        # Look for the header separator
        if re.match(r'^-+$', line.strip()):
            in_data_section = True
            continue

        if not in_data_section:
            continue

        # Skip empty lines
        if not line.strip():
            continue

        # Parse data lines
        # Format: endpoint_name  required_delay  actual_delay  slack  (status)
        # Example: _42791_/D (sky130_fd_sc_hs__dfrtp_1)    9.8396   12.9553   -3.1157 (VIOLATED)
        match = re.match(
                r'^(\S+)\s+\(([^)]+)\)\s+([\d.]+)\s+([\d.]+)\s+([-\d.]+)(?:\s+\((\w+)\))?',
                line
        )

        if match:
            endpoint = match.group(1)
            cell_type = match.group(2)
            required_delay = float(match.group(3))
            actual_delay = float(match.group(4))
            slack = float(match.group(5))
            status = match.group(6) if match.group(6) else "MET"

            paths.append({
                'endpoint': endpoint,
                'cell_type': cell_type,
                'required': required_delay,
                'arrival': actual_delay,
                'slack': slack,
                'status': status
            })

    return paths


def calculate_metrics(paths: list, top_module: str):
    """Calculate timing metrics from parsed paths."""

    total_paths = len(paths)

    if total_paths == 0:
        return {
            'top_module': top_module,
            'total_paths': 0,
            'wns': 0.0,
            'tns': 0.0,
            'nfe': 0,
            'sum_arrival_required_ratio': 0.0,
            'slack_list': []
        }

    # Initialize metrics
    wns = float('inf')  # Worst Negative Slack (minimum slack)
    tns = 0.0           # Total Negative Slack
    nfe = 0             # Number of Failing Endpoints
    sum_arrival_required_ratio = 0.0
    slack_list = []

    for path in paths:
        slack = path['slack']
        slack_list.append(slack)

        # Update WNS (minimum slack)
        wns = min(wns, slack)

        # Update TNS and NFE for negative slack
        if slack < 0:
            tns += slack
            nfe += 1

        # Calculate arrival/required ratio
        required = path['required']
        arrival = path['arrival']

        if required != 0:
            ratio = arrival / required
            sum_arrival_required_ratio += ratio

    # If no negative slack found, WNS should be the best (least negative or most positive)
    if wns == float('inf'):
        wns = min(slack_list) if slack_list else 0.0

    return {
        'top_module': top_module,
        'total_paths': total_paths,
        'wns': wns,
        'tns': tns,
        'nfe': nfe,
        'sum_arrival_required_ratio': sum_arrival_required_ratio,
        'slack_list': slack_list
    }


def write_csv(metrics: dict, output_file: str) -> None:
    """Write metrics to CSV file."""

    with open(output_file, 'w', encoding='utf=8', newline='') as f:
        writer = csv.writer(f)

        # Write header
        header = ['Top_Module', 'Total_Paths', 'WNS', 'TNS', 'NFE',
                  'Sum_Arrival_Required_Ratio', 'Path_Slacks']
        writer.writerow(header)

        # Write data row
        row = [
            metrics['top_module'],
            metrics['total_paths'],
            f"{metrics['wns']:.4f}",
            f"{metrics['tns']:.4f}",
            metrics['nfe'],
            f"{metrics['sum_arrival_required_ratio']:.4f}"
        ]

        # Append all slack values
        for slack in metrics['slack_list']:
            row.append(f"{slack:.4f}")

        writer.writerow(row)

# pylint: disable=too-many-locals,too-many-branches,too-many-statements
def print_slack_histogram(
        slack_list: list, bin_width = None, paths_per_char: int = 1
) -> None:
    """
    Print an ASCII-art histogram of slack distribution.

    Args:
        slack_list: List of slack values
        bin_width: Width of each bin (auto-calculated if None)
        paths_per_char: Number of paths represented by each character
    """

    if not slack_list:
        print("\nNo paths to display in histogram.")
        return

    # Calculate the range and ensure 0 is a bin boundary
    min_slack = min(slack_list)
    max_slack = max(slack_list)

    # If bin_width not specified, calculate a reasonable default
    # Target around 20-30 total bins
    if bin_width is None:
        total_range = max(abs(min_slack), abs(max_slack)) * 2
        bin_width = total_range / 25  # Aim for ~25 bins

        # Round to a nice number for readability
        magnitude = 10 ** (len(str(int(bin_width))) - 1)
        bin_width = max(0.1, round(bin_width / magnitude) * magnitude)

    # Build bin edges with 0 as an exact boundary
    bin_edges = []

    # Create negative bins (from min_slack up to 0)
    if min_slack < 0:
        # Calculate how many bins we need to cover the negative range
        num_neg_bins = int(abs(min_slack) / bin_width) + 1
        # Start from a multiple of bin_width below min_slack
        neg_start = -num_neg_bins * bin_width

        for i in range(num_neg_bins):
            edge = neg_start + i * bin_width
            if edge >= min_slack:
                bin_edges.append(edge)

    # Add zero boundary
    bin_edges.append(0.0)

    # Create positive bins (from 0 up to max_slack)
    if max_slack > 0:
        # Calculate how many bins we need to cover the positive range
        num_pos_bins = int(max_slack / bin_width) + 1

        for i in range(1, num_pos_bins + 1):
            edge = i * bin_width
            bin_edges.append(edge)
            if edge >= max_slack:
                break

    # Remove duplicates and sort (in case of floating point issues)
    bin_edges = sorted(list(set(bin_edges)))

    # Initialize bins
    num_bins = len(bin_edges) - 1
    bins = [0] * num_bins

    # Count paths in each bin
    for slack in slack_list:
        # Find the appropriate bin
        for i in range(num_bins):
            if bin_edges[i] <= slack < bin_edges[i + 1]:
                bins[i] += 1
                break
        else:
            # Handle the edge case where slack equals the maximum
            if slack == bin_edges[-1]:
                bins[-1] += 1

    # Print histogram
    print("\n" + "=" * 70)
    print("Slack Distribution Histogram")
    print("=" * 70)
    print(f"Total Paths: {len(slack_list)}")
    print(f"Slack Range: [{min_slack:.4f}, {max_slack:.4f}]")
    print(f"Uniform Bin Width: {bin_width:.4f}")
    print(f"Each character represents {paths_per_char} path(s)")
    print()

    # Print header
    print(f"{'Slack Range':<25} {'Count':<8} Distribution")
    print("-" * 70)

    # Print each bin
    for i in range(num_bins):
        bin_start = bin_edges[i]
        bin_end = bin_edges[i + 1]
        count = bins[i]

        # Calculate bar width based on paths_per_char
        bar_width = count // paths_per_char

        # Choose character based on whether slack is negative (violated)
        if bin_end <= 0:
            bar_char = 'X'  # Violated paths
        else:
            bar_char = '='  # Paths meeting timing

        # Create bar
        _bar = bar_char * bar_width

        # Format bin range
        bin_label = f"[{bin_start:>7.4f}, {bin_end:>7.4f})"

        # Print row
        print(f"{bin_label:<25} {count:<8} {_bar}")

    print("-" * 70)
    print("Legend: X = Violated (slack < 0), = = Met (slack >= 0)")
    print("=" * 70)


def print_summary(metrics):
    """Print timing metrics summary to console."""

    print("\n" + "=" * 42)
    print("Timing Metrics Summary")
    print("=" * 42)
    print(f"Top Module:                    {metrics['top_module']}")
    print(f"Total Paths to Registers:      {metrics['total_paths']}")
    print(f"Worst Negative Slack (WNS):    {metrics['wns']:.4f}")
    print(f"Total Negative Slack (TNS):    {metrics['tns']:.4f}")
    print(f"Number of Failing Endpoints:   {metrics['nfe']}")
    print(f"Sum Arrival/Required Ratio:    {metrics['sum_arrival_required_ratio']:.4f}")
    print("=" * 42)


def main():
    """Main function."""

    # Get environment variables or use defaults
    timing_report_file = os.getenv('TIMING_REPORT_FILE', 'timing_report.txt')
    top_module = os.getenv('TOP_MODULE', 'unknown')
    output_csv = os.getenv('OUTPUT_CSV', 'timing_metrics.csv')

    # Allow command line override
    if len(sys.argv) > 1:
        timing_report_file = sys.argv[1]
    if len(sys.argv) > 2:
        top_module = sys.argv[2]
    if len(sys.argv) > 3:
        output_csv = sys.argv[3]

    # Check if report file exists
    if not os.path.exists(timing_report_file):
        print(f"Error: Timing report file '{timing_report_file}' not found.", file=sys.stderr)
        sys.exit(1)

    print(f"Processing timing report: {timing_report_file}")

    # Parse timing report
    paths = parse_timing_report(timing_report_file)

    # Calculate metrics
    metrics = calculate_metrics(paths, top_module)

    # Write CSV output
    write_csv(metrics, output_csv)
    print(f"CSV file written to: {output_csv}")

    # Print summary
    print_summary(metrics)

    # Print slack histogram
    print_slack_histogram(metrics['slack_list'])


if __name__ == '__main__':
    main()



'''Helper for adding source files:

Allows user to explicitly add systemverilog files that don't end in .sv using

sv@my_module.txt
v@my_verilog_module.txt
vhdl@my_vhdl_code.log

Otherwise eda.py can't know if a .txt (or .pcap, etc) file is a source file
to be part of compilation, or a file needed for simulation (.txt, .pcap, .mem
as part of a verilog $readmemh, etc)

'''

import os
from pathlib import Path
import shutil

# Ways to force files not ending in .sv to be systemverilog (for tools
# that require -sv vs Verilog-2001'''
FORCE_PREFIX_DICT = {
    # The values must match what's expected by eda.CommandDesign.add_file,
    # which are named in eda_config_defaults.yml - file_extensions:
    'sv@': 'systemverilog',
    'v@': 'verilog',
    'vhdl@': 'vhdl',
    'cpp@': 'cpp',
    'sdc@': 'synth_constraints',
    'f@': 'dotf',
    'py@' : 'python',
    'makefile@': 'makefile'
}

ALL_FORCED_PREFIXES = set(list(FORCE_PREFIX_DICT.keys()))

def get_source_file(target: str) -> (bool, str, str):
    '''Returns tuple: bool if file exists, filepath str, and optional forced file type str'''

    if '$' in target:
        target = os.path.expandvars(target)

    if os.path.isfile(target):
        # target exists as a file, return True w/ original target:
        return True, target, ''

    if '@' in target:
        for p in ALL_FORCED_PREFIXES:
            if p in target:
                # essentially removing the leading "sv@" or whatever prefix.
                fpath = ''.join(target.split(p))
                if os.path.isfile(fpath):
                    return True, fpath, FORCE_PREFIX_DICT.get(p)

    # target or fpath didn't exist, return False with the original target:
    return False, target, ''


def safe_shutil_which(path: str) -> str:
    '''Windows/WSL compatible Wrapper for shutil.which that checks for 'path':

    - path
    - path.exe
    - path.bat

    Returns full path str returned by shutil.which
    '''
    for ext in ('', '.exe', 'bat'):
        if found := shutil.which(f'{path}{ext}'):
            return found
    return ''


def get_source_files_paths(
        cmd_des_obj, minimal_root_paths: bool = False
) -> list:
    '''Given a CommandDesign object, returns list of Path objs for all source files

    If minimal_root_paths=True, attempts to return the minimal set of root paths for
    docker read-only volume mapping

    will look at members for all source files that (any member list type in cmd_des_obj.files_)
    '''

    # Iterate over lists and add to our set, we'll include cmd_des_obj.incdirs (list)
    # here as well.
    file_dirs = set()
    for name in [x for x in dir(cmd_des_obj) if x.startswith('files_')]:
        member = getattr(cmd_des_obj, name, None)
        if callable(member) or not isinstance(member, list):
            continue
        _ = [file_dirs.add(Path(x).resolve().parent) for x in member]

    # also incdirs, a path not a file, so don't use .parent
    _ = [file_dirs.add(Path(incdir).resolve()) for incdir in getattr(cmd_des_obj, 'incdirs', [])]

    if not minimal_root_paths:
        # return them all:
        return list(file_dirs)

    for x in list(file_dirs):
        # - Using list(file_dirs) as iterator as cheap copy b/c we're altering set file_dirs
        # - We don't want to add '/' as a root path, so make sure dirs have at least
        #   2 parts if we're going to trim items from file_dirs.
        # - Also this is a n^2 traversal, make sure we aren't matching ourselves (x != y)
        #   in check. Using childPath.is_relative_to(parentPath).
        if any(x != y and x.is_relative_to(y) and len(y.parts) >= 2 for y in file_dirs):
            file_dirs.remove(x)

    return list(file_dirs)


# Note that in Windows, 'python' will return the venv or uv version, 'python3' will
# return the installed version (which may not be what you want), so we'll prefer
# 'python':
PY_EXE = safe_shutil_which('python') or safe_shutil_which('python3')


def safe_output_file(work_dir: str, default_fname: str, user_fpath: str = '') -> str:
    '''Wrapper for getting a safe'ish abspath output file name, often from user supplied args

    For example:
      - user_fpath='', will use (abspath) work_dir / default_fname
      - user_fpath='results.txt', will use (abspath) work_dir prepended
      - user_fpath='./results.txt', will use (abspath) ./results.txt
           -- with safely creating the user supplied directory
    '''

    if user_fpath:
        outpath, outleaf = os.path.split(user_fpath)
    else:
        outpath, outleaf = '', ''

    assert default_fname

    if not outpath and not outleaf:
        # Use default.
        output_dir = os.path.abspath(work_dir)
        output_file = os.path.join(output_dir, default_fname)
    elif not outpath:
        output_dir = os.path.abspath(work_dir)
        output_file = os.path.join(output_dir, outleaf)
    else:
        # User supplied path + file
        output_dir = os.path.abspath(outpath)
        os.makedirs(output_dir, exist_ok=True)
        output_file = os.path.join(output_dir, outleaf)

    return output_file



'''opencos.deps.deps_commands - pymodule for processing DEPS markup deps-commands

This supports things like:
my_target:
  - deps:
    - commands:
      - shell: <string>
      - peakrdl: <string>
      - work-dir-add-sources: <list>
'''


import os
from pathlib import Path
import re

from opencos.deps.defaults import SUPPORTED_COMMAND_KEYS, COMMAND_ATTRIBUTES
from opencos.files import safe_shutil_which, PY_EXE
from opencos.util import debug, error, warning, ShellCommandList


THISPATH = os.path.dirname(__file__)
PEAKRDL_CLEANUP_PY = os.path.join(THISPATH, '..', 'peakrdl_cleanup.py')

def path_substitutions_relative_to_work_dir(
        exec_list: list, info_str: str, target_path: str,
        attributes: dict
) -> list:
    '''For shell commands, since eda.py Command objects operate out of a "work-dir", it can
    make shell commands confusing with additional ../.. to get to the calling directory, and
    the original target directory is completely lost otherwise.

    By default, if a FILE exists in the calling target's path, we will substitute that file
    instead of looking in the work-dir. (can be disabled via filepath-subst-target-dir: False)

    Optionally if a DIR exists in the calling target's path, will will substitute that DIR
    instead of relative to work-dir. This can be annoying with relative paths like ../ (which
    would exist in both) so this is disabled by default (can be enabled with
    dirpath-subst-target-dir: True)
    '''

    if not attributes.get('filepath-subst-target-dir', True) and \
       not attributes.get('dirpath-subst-target-dir', False):
        return exec_list

    # Look for path substitutions, b/c we later "work" in self.args['work-dir'], but
    # files should be relative to our target_path.
    for i,word in enumerate(exec_list):
        m = re.search(r'(\.+\/+[^"\;\:\|\<\>\*]*)$', word)
        if m:
            # ./, ../, file=./../whatever  It might be a filepath.
            # [^"\;\:\|\<\>\*] is looking for non-path like characters, so we dont' have a trailing
            #  " : ; < > |
            # try and see if this file or dir exists. Note that files in the
            # self.args['work-dir'] don't need this, and we can't assume dir levels in the work-dir.
            try:
                try_path = os.path.abspath(os.path.join(os.path.abspath(target_path), m.group(1)))
                if attributes.get('filepath-subst-target-dir', True) and \
                   os.path.isfile(try_path):
                    # make the substitution
                    exec_list[i] = word.replace(m.group(1), try_path)
                    debug(f'file path substitution {info_str=} {target_path=}: replaced - {word=}'
                          f'is now ={exec_list[i]}. This can be disabled in DEPS with:',
                          '"filepath-subst-targetdir: false"')
                elif attributes.get('dirpath-subst-target-dir', False) and \
                     os.path.isdir(try_path):
                    # make the substitution
                    exec_list[i] = word.replace(m.group(1), try_path)
                    debug(f'dir path substitution {info_str=} {target_path=}: replaced - {word=}'
                          f'is now ={exec_list[i]}. This can be disabled in DEPS with:',
                          '"dirpath-subst-targetdir: false"')
            except Exception:
                pass

    return exec_list


def line_with_var_subst( # pylint: disable=dangerous-default-value
        line : str, replace_vars_dict: dict = {},
        replace_vars_os_env: bool = False,
        target_node: str = '', target_path: str = ''
) -> str:
    '''Given a line (often from a shell style command) perform var substitution'''

    # We can try for replacing any formatted strings, using self.args, and os.environ?
    # We have to do this per-word, so that missing replacements or tcl-like things, such
    # as '{}' wouldn't bail if trying to do line.format(**dict)
    if '{' not in line:
        return line

    replace_dict = replace_vars_dict

    words = line.split()
    for i,word in enumerate(words):
        try:
            words[i] = word.format(**replace_dict)
        except Exception:
            # this was an attempt, no need to catch exceptions if we couldn't replace the word.
            pass

    new_line = ' '.join(words)

    if replace_vars_os_env:
        try:
            new_line = os.path.expandvars(new_line)
        except Exception as e:
            warning(f'{e}: trying to apply env vars to line: {new_line}')

    if new_line != line:
        debug(f'{target_node=} {target_path=} performed string format replacement,',
              f'{line=} {new_line=}')
        return new_line

    debug(f'{target_node=} {target_path=} string format replacement attempted,',
          f'no replacement. {line=}')
    return line


def parse_deps_shell_str(
        line: str, target_path: str, target_node: str,
        attributes: dict,
        enable: bool = True
) -> dict:
    '''Returns None or a dict of a possible shell command from line (str)

     Examples of 'line' str:
         shell@echo "hello world" > hello.txt
         shell@ generate_something.sh
         shell@ generate_this.py --input=some_data.json
         shell@ cp ./some_file.txt some_file_COPY.txt
         shell@ vivado -mode tcl -script ./some.tcl -tclargs foo_ip 1.2 foo_part foo_our_name \
                      {property value}

    Returns None if no parsing was performed, or if enable is False

    target_path (str) -- from dependency parsing (relative path of the DEPS file)
    target_node (str) -- from dependency parsing, the target containing this 'line' str.
    '''
    if not enable:
        return {}

    m = re.match(r'^\s*shell\@(.*)\s*$', line)
    if not m:
        return {}

    exec_str = m.group(1)
    exec_list = exec_str.split()

    # Look for path substitutions, b/c we later "work" in self.args['work-dir'], but
    # files should be relative to our target_path.
    # Note this can be disable in DEPS via path-subst-target-dir=False
    exec_list = path_substitutions_relative_to_work_dir(
        exec_list=exec_list, info_str='shell@', target_path=target_path,
        attributes=attributes
    )

    return {
        'target_path': os.path.abspath(target_path),
        'target_node': target_node,
        'exec_list': exec_list,
        'attributes': attributes
    }


def parse_deps_work_dir_add_srcs(
        line : str, target_path : str, target_node : str,
        attributes: dict,
        enable : bool = True
) -> dict:
    '''Returns None or a dict describing source files to add from the work-dir path

     Examples of 'line' str:
         work_dir_add_srcs@ my_csrs.sv
         work_dir_add_srcs@ some_generated_file.sv some_dir/some_other.v ./gen-vhd-dir/even_more.vhd

    Returns None if no parsing was performed, or if enable is False

    target_path (str) -- from dependency parsing (relative path of the DEPS file)
    target_node (str) -- from dependency parsing, the target containing this 'line' str.
    '''
    if not enable:
        return {}

    m = re.match(r'^\s*work_dir_add_srcs\@(.*)\s*$', line)
    if not m:
        return {}

    files_str = m.group(1)
    file_list = files_str.split()

    d = {
        'target_path': os.path.abspath(target_path),
        'target_node': target_node,
        'file_list': file_list,
        'attributes': attributes
    }
    return d


def parse_deps_peakrdl( # pylint: disable=too-many-locals
        line: str, target_path: str, target_node: str,
        attributes: dict,
        enable: bool = True
) -> dict:
    '''Returns None or a dict describing a PeakRDL CSR register generator dependency

     Examples of 'line' str:
         peakrdl@ --cpuif axi4-lite-flat --top oc_eth_10g_1port_csrs ./oc_eth_10g_csrs.rdl

    Returns None if no parsing was performed, or if enable=False

    target_path (str) -- from dependency parsing (relative path of the DEPS file)
    target_node (str) -- from dependency parsing, the target containing this 'line' str.
    '''

    m = re.match(r'^\s*peakrdl\@(.*)\s*$', line)
    if not m:
        return None

    if not enable:
        warning(f'peakrdl: encountered peakrdl command in {target_path=} {target_node=},' \
                + ' however it is not enabled in edy.py - eda.config[dep_command_enables]')
        return None

    if not safe_shutil_which('peakrdl'):
        error('peakrdl: is not present in shell path, or the python package is not avaiable,' \
              + f' yet we encountered a peakrdl command in {target_path=} {target_node=}')
        return None


    args_str = m.group(1)
    args_list = args_str.split()

    # Fish out the .rdl name
    # If there is --top=value or --top value, then fish out that value (that will be the
    # value.sv and value_pkg.sv generated names.

    sv_files = []
    top = ''
    for i,str_value in enumerate(args_list):
        if '--top=' in str_value:
            _, top = str_value.split('=')
        elif '--top' in str_value:
            if i + 1 < len(args_list):
                top = args_list[i + 1]

    for str_item in args_list:
        if str_item[-4:] == '.rdl':
            _, rdl_fileonly = os.path.split(str_item) # strip all path info
            rdl_filebase, _ = os.path.splitext(rdl_fileonly) # strip .rdl ext
            if not top:
                top = rdl_filebase

    assert top != '', \
        f'peakrdl@ DEP, could not determine value {top=}: {line=}, {target_path=}, {target_node=}'

    sv_files += [ f'peakrdl/{top}_pkg.sv', f'peakrdl/{top}.sv' ]

    shell_commands = [
        [ safe_shutil_which('peakrdl'), 'regblock', '-o', str(Path('peakrdl/'))] + args_list,
        # Edit file to apply some verilator waivers, etc, from peakrdl_cleanup.py:
        [ PY_EXE, PEAKRDL_CLEANUP_PY, str(Path(f'peakrdl/{top}.sv')),
          str(Path(f'peakrdl/{top}.sv')) ],
    ]

    ret_dict = {
        'shell_commands_list': [], # Entry needs target_path, target_node, exec_list
        'work_dir_add_srcs': {},   # Single dict needs target_path, target_node, file_list
    }

    # Make these look like a dep_shell_command:
    for one_cmd_as_list in shell_commands:
        ret_dict['shell_commands_list'].append(
            parse_deps_shell_str(
                line=(' shell@ ' + ' '.join(one_cmd_as_list)),
                target_path=target_path,
                target_node=target_node,
                attributes=attributes
            )
        )

    # Make the work_dir_add_srcs dict:
    ret_dict['work_dir_add_srcs'] = parse_deps_work_dir_add_srcs(
        line=' work_dir_add_srcs@ ' + ' '.join(sv_files),
        target_path=target_path,
        target_node=target_node,
        attributes=attributes
    )

    return ret_dict



def deps_commands_handler( #pylint: disable=too-many-locals,too-many-branches
        config: dict, eda_args: dict,
        dep : str, deps_file : str, target_node : str, target_path : str,
        commands : list
) -> (list, list):
    ''' Returns a tuple of (shell_commands_list, work_dir_add_srcs_list), from processing
        a DEPS.yml entry for something like:

        target_foo:
          deps:
            - some_file
            - commands: # (list of dicts) These are directly in a 'deps' list.
                - shell: ...
                - peakrdl: ...
                - work-dir-add-sources: ...
                - shell: ...

        target_foo:
          commands: # (list of dicts) These are in a target, but not ordered with other deps
             - shell: ...
             - peakrdl: ...
             - work-dir-add-sources: ...
             - shell: ...

        We'd like to handle the list in a 'commands' entry, supporting it in a few places in a
        DEPS.yml, so this this a generic way to do that. Currently these are broken down into Shell
        commands and Files that will be later added to our sources (b/c we haven't run the Shell
        commands yet, and the Files aren't present yet but we'd like them in our eda.py filelist in
        order.

    '''

    shell_commands_list = []
    work_dir_add_srcs_list = []

    for command in commands:
        assert isinstance(command, dict), \
            (f'{type(command)=} must be dict, for {deps_file=} {target_node=} {target_path=} with'
             f'{commands=}')

        for key in command.keys():
            if key not in SUPPORTED_COMMAND_KEYS:
                error(f'deps_commands.process_commands - command {key=} not in',
                      f'{SUPPORTED_COMMAND_KEYS=}')

        # collect attributes for this command:
        attributes = {}
        for key, default_value in COMMAND_ATTRIBUTES.items():
            attributes[key] = command.get(key, default_value)

        var_subst_dict = {} # this is per-command.
        if config['dep_command_enables'].get('var_subst_args', False) and \
           attributes['var-subst-args']:
            var_subst_dict = eda_args

        for key,item in command.items():

            # skip the tee and var-subst-* keys, since these types are bools and not commands.
            if key in COMMAND_ATTRIBUTES:
                continue

            # Optional variable substituion in commands
            if isinstance(item, str):
                item = line_with_var_subst(
                    line=item,
                    replace_vars_dict=var_subst_dict,
                    replace_vars_os_env=attributes['var-subst-os-env'],
                    target_node=target_node,
                    target_path=deps_file
                )

            if key == 'shell':
                # For now, piggyback on parse_deps_shell_str:

                # If our shell commands have \n in them, split the commands.
                item = item.replace('\\\n', ' ') # but not \ + newline

                for shell_line in item.split('\n'):

                    shell_line = shell_line.strip()
                    if not shell_line:
                        continue

                    ret_dict = parse_deps_shell_str(
                        line=('shell@ ' + shell_line),
                        target_path=target_path,
                        target_node=target_node,
                        attributes=attributes,
                        enable=config['dep_command_enables']['shell'],
                    )
                    # To support 'tee: <some-file>' need to append it to last
                    # list item in ret_dict['exec_list'], and make it a util.ShellCommandList.
                    if attributes['tee']:
                        ret_dict['exec_list'] = ShellCommandList(
                            ret_dict['exec_list'], tee_fpath=attributes['tee']
                        )
                    assert ret_dict, \
                        f'shell command failed in {dep=} {target_node=} in {deps_file=}'
                    shell_commands_list.append(ret_dict) # process this later, append to ret tuple

            elif key in ['work-dir-add-srcs', 'work-dir-add-sources']:
                # For now, piggyback on parse_deps_work_dir_add_srcs:
                ret_dict = parse_deps_work_dir_add_srcs(
                    line='work_dir_add_srcs@ ' + item,
                    target_path=target_path,
                    target_node=target_node,
                    enable=config['dep_command_enables']['work_dir_add_srcs'],
                    attributes=attributes,
                )
                assert ret_dict, \
                    f'work-dir-add-srcs command failed in {dep=} {target_node=} in {deps_file=}'

                work_dir_add_srcs_list.append(ret_dict) # process this later, append to ret tuple

            elif key == 'peakrdl':
                # for now, piggyback on parse_deps_peakrdl:
                ret_dict = parse_deps_peakrdl(
                    line='peakrdl@ ' + item,
                    target_path=target_path,
                    target_node=target_node,
                    enable=config['dep_command_enables']['peakrdl'],
                    attributes=attributes
                )
                assert ret_dict, f'peakrdl command failed in {dep=} {target_node=} in {deps_file=}'

                # add all the shell commands:
                shell_commands_list += ret_dict['shell_commands_list'] # several entries.
                # all the work_dir_add_srcs:
                work_dir_add_srcs_list += [ ret_dict['work_dir_add_srcs'] ] # single entry append


            else:
                assert False, \
                    f'unknown {key=} in {command=}, {item=} {dep=} {target_node=} in {deps_file=}'

    return (shell_commands_list, work_dir_add_srcs_list)



'''Handler for: eda waves --tool=cognichip or --tool=cognichip-beta'''

import json
from pathlib import Path
import os

from opencos.commands.waves import CommandWaves
from opencos.eda_base import Tool
from opencos.files import safe_shutil_which
from opencos.utils.subprocess_helpers import IS_LINUX, IS_MACOS, IS_WINDOWS
from opencos.utils.macos_utils import get_macos_tool_info
from opencos.utils.win_utils import get_windows_tool_info

class ToolCognichip(Tool):
    '''
    Derived class for tool specifics for Cognichip

    Holds version, and waveviewer_exe
    '''

    _TOOL = 'cognichip'
    _EXE = 'cognichip'

    waveviewer_exe = ''

    def __init__(self, config: dict):
        super().__init__(config=config)

    def get_versions(self, **kwargs) -> str:
        if self._VERSION:
            return self._VERSION

        self._VERSION = ''

        # cognichip[-beta] --version is fast enough to call subprocess, but that
        # doesn't return the correct value (yet), so instead look for the
        # (note: look for correct 'cognichip' vs 'cognichip-beta'):
        #   Linux: EXE_PATH/../share/cognichip-beta/resources/app/product.json
        #   Macos: /Applications/Cognichip Beta.app/Contents/Resources/app/product.json
        #   Win: $env:LOCALAPPDATA\Programs\Cognichip Beta\resources\app\product.json
        # and look for key "cognichipVersion".
        self.waveviewer_exe = safe_shutil_which(self._EXE)

        if IS_LINUX:
            # For macos, shutil which won't work for cognichip, we'll check later.
            # For windows, we could require it to be in $env:Path, but we can also
            # fish it out of $env:LOCALPPDATA if not in Path.
            if not self.waveviewer_exe:
                return self._VERSION

        product_json_fpath = ''

        if IS_LINUX:
            bin_fpath = Path(self.waveviewer_exe).parent
            product_json_fpath = (
                bin_fpath.parent / 'share' / self._TOOL / 'resources' / 'app' / 'product.json'
            )

        elif IS_MACOS:
            # Note that eda_tool_helper and eda have to be aware of macos tax.
            app_name = self.config.get('tools', {}).get(self._TOOL, {}).get(
                'macos-app-name', ''
            )
            if app_name:
                data = get_macos_tool_info(
                    app_name, prefer_exes=[
                        self._EXE, 'cognichip', 'code'
                    ]
                )
                product_json_fpath = Path(data.get('metadata', ''))
                if exe := data.get('bin', ''):
                    # pass back to what macos tells us:
                    self.waveviewer_exe = exe
                    self._EXE = exe

        elif IS_WINDOWS:
            app_name = 'Cognichip'
            data = get_windows_tool_info(app_name, prefer_exes=[self._EXE])
            product_json_fpath = Path(data.get('metadata', ''))
            if exe := data.get('bin', ''):
                # pass back to what win_utils tells us:
                self.waveviewer_exe = exe
                self._EXE = exe


        if not product_json_fpath or not product_json_fpath.is_file():
            return self._VERSION

        data = {}
        with open(str(product_json_fpath), encoding='utf-8') as f:
            data = json.load(f)
        self._VERSION = data.get('cognichipVersion', '')
        return self._VERSION


class CommandWavesCognichip(CommandWaves, ToolCognichip):
    '''Handler for: eda waves --tool=cognichip

    This is registerd in eda's CONFIG-YML config.tools.cognichip.handlers.waves
    '''

    def __init__(self, config: dict):
        CommandWaves.__init__(self, config=config)
        ToolCognichip.__init__(self, config=self.config)

    def do_it(self) -> None:
        command_list = [
            self.waveviewer_exe, '--disable-workspace-trust', '-r', self.wave_file
        ]
        self.exec(os.path.dirname(self.wave_file), command_list)



'''
Helper to resolve simply finding an exe, across OS differences
'''


from opencos import eda_config
from opencos.files import safe_shutil_which
from opencos.utils import macos_utils, vsim_helper, win_utils
from opencos.utils.subprocess_helpers import IS_MACOS, IS_WINDOWS


def get_tool_exe(config: dict, tool: str) -> str:
    '''Returns an exe that can be used for this tool'''

    # Use the macos_utils to resolve that exe.
    # If this is mac, and the tool (in config.tools.TOOLNAME) has
    # macos-app-name defined, then we have to query that.
    # otherwise, we get the first exe listed from config.tools.TOOLNAME
    if IS_MACOS:
        exe = macos_utils.get_tool_exe_from_config(config, tool)

    elif IS_WINDOWS:
        # Note that WSL will not report IS_WINDOWS, will report IS_LINUX,
        # so we don't go down this path (looking in C:\ for LOCALAPPDATA AND
        # PROGRAMFILES) for WSL.
        exe = win_utils.get_tool_exe_from_config(config, tool)

    else:
        tool_cfg = config.get('tools', {}).get(tool, {})
        exe_list = eda_config.tool_get_exe_list(tool=tool, config=config)
        exe = exe_list[0] # pick first one, or tool name


    # Attempt to resolve, compatible for IS_LINUX, IS_WINDOWS,
    # or IS_WINDOWS+WSL:
    exe = safe_shutil_which(exe)

    # special hook for eda_config - tools - TOOLNAME - requires_vsim_helper:
    tool_cfg = config.get('tools', {}).get(tool, {})
    if tool_cfg.get('requires-vsim-helper', False):
        exe = vsim_helper.lookup_exe(tool)

    return exe
